/**
 * @file motor.cpp
 * @author Keith Holman
 * @date Created: Tuesday, October 14, 2008
 * @date Created: Sunday, November 30, 2008
 *
 * @brief Defines the atomic model for the motor
 */

#include "motor.h"
#include "message.h"    // class ExternalMessage, InternalMessage
#include "mainsimu.h"   // MainSimulator::Instance().getParameter( ... )
#include "NXT++.h"      // For controlling the NXT motor
#include <cmath>
#include <iostream>

Motor::Motor( const std::string& name )
: BaseMotor( name ),
  _in( addInputPort( "in" ) ),
  _out( addOutputPort( "out" ) ),
  _circumference( 0.0 ),
  _expelled( 0.0 ),
  _target( 0.0 ),
  _total( 100.0 ),
  _success( true ),
  _rotationsPerSecond( 0.0 )
{
  const static double PI = 4.0 * atan( 1.0 );
  _circumference = PI * 
    atof( MainSimulator::Instance().getParameter(
      description(), "diameter" ).c_str() );
  _expelled = 
    atof( MainSimulator::Instance().getParameter(
      description(), "expelled" ).c_str() );
  _total =
    atof( MainSimulator::Instance().getParameter(
      description(), "length" ).c_str() );

  _rotationsPerSecond =
    atof( MainSimulator::Instance().getParameter(
      description(), "rotationsPerSecond" ).c_str() );
  /* Prevent weird inputs */
  if ( _circumference <= 0.0 ) _circumference = 1.0;
  if ( _rotationsPerSecond <= 0.0 ) _rotationsPerSecond = 1.0;
}

Motor::~Motor()
{
}

Model &Motor::initFunction()
{
  this->passivate();
  return *this;
}

Model &Motor::externalFunction( const ExternalMessage &msg )
{
  if( msg.port() == _in && state() == passive )
  {
    Value len = msg.value();
    Value delta = 0.0;
    _success = len > 0 && len < _total;
    if ( _success )
    {
      delta = len - _expelled;
      _target = len;
    }
    else
    {
      if ( len < 0.0 )
      {
        delta = _expelled;
        _target = 0;
      }
      else
      {
        delta = _total - _expelled;
        _target = _total;
      }
    }
    std::cout << "DELTA: " << delta << std::endl;
    bool forward = ( delta >= 0.0 ) ? true : false;
    delta = fabs( delta );
    Value numberOfRotations = delta / _circumference;
    Value val_sec = numberOfRotations / _rotationsPerSecond;

    /* Calculate how long the motor needs to be in a moving state */
    unsigned int min = static_cast< unsigned int >( val_sec / 60.0 );
    unsigned int sec = static_cast< unsigned int >( val_sec ) % 60;
    unsigned int msec = static_cast< unsigned int >(
      ceil( ( val_sec - static_cast< int >( val_sec ) ) * 1000 ) );

    NXT::Motor::ResetRotationCount( _port, false );
    if ( forward )
      NXT::Motor::SetForward( _port, _motorPower );
    else
      NXT::Motor::SetReverse( _port, _motorPower );
    std::cout << "min: " << min << " sec: " << sec << " msec: " << msec << std::endl;
    Time t( 0, min, sec, msec );
    holdIn( active, t );
  }
  return *this;
}

Model &Motor::internalFunction( const InternalMessage & )
{
  _expelled = _target;
  passivate();
  return *this;
}

Model &Motor::outputFunction( const InternalMessage &msg )
{
  NXT::Motor::Stop( _port, true );
  sendOutput( msg.time(), _out, _success );
  return *this ;
}

std::string Motor::className() const
{
  return "Motor";
}
