/**
 * @file planner.cpp
 * @author Keith Holman
 * @date Created: Tuesday, October 14, 2008
 *
 * @brief Defines the atomic model for the planner
 */

#include "planner.h"
#include "message.h"    // class ExternalMessage, InternalMessage
#include "mainsimu.h"   // MainSimulator::Instance().getParameter( ... )
#include "compare.h"    // for is_not_equal

#include <cstdlib>

Value Planner::STEP_SIZE = 1.25;
Value Planner::NEGATIVE_STEP_SIZE = -1.25;

Planner::Planner( const std::string& name )
: Atomic( name ),
  _in( addInputPort( "in" ) ),
  _done( addInputPort( "done" ) ),
	_out( addOutputPort( "out" ) ),
	_ready( addOutputPort( "ready" ) ),
	_engaged( addOutputPort( "engaged" ) ),
  _nextPosition( addOutputPort( "nextPosition" ) ),
  _success( true ),
  _outputSuccess( false ),
  _visible( false ),
  _lastVisibility( false ),
  _position(),
  _destination()
{
  _position.x( atof( MainSimulator::Instance().getParameter( description(), "intial.x" ).c_str() ) );
  _position.y( atof( MainSimulator::Instance().getParameter( description(), "intial.y" ).c_str() ) );
  _destination = _position;
}

Model &Planner::initFunction()
{
	this->passivate();
	return *this;
}

Model &Planner::externalFunction( const ExternalMessage &msg )
{
	if ( msg.port() == _in )
	{
	  _inputData.put( msg.value() );
	  if ( _inputData.full() )
	  {
	    _visible = _inputData.get(VISIBLE);
	    _destination = Point( _inputData.get(X_COORD),
                          _inputData.get(Y_COORD) );
      std::cout << "X: " << _destination.x() << ", Y: " << _destination.y() 
                << std::endl;
	    _inputData.clear();
	    _success = true;
	  }
	}
  else if ( msg.port() == _done )
  {
    if ( _success )
      _success = static_cast< bool >( msg.value() );
    _outputSuccess = true;
  }
	holdIn( active, Time( 0, 0, 0, 0 ) );
	return *this;
}

Model &Planner::internalFunction( const InternalMessage & )
{
  passivate();
	return *this;
}

Model &Planner::outputFunction( const InternalMessage &msg )
{
  static unsigned int count = 0;
  if ( fabs( _position.x() - _destination.x() ) >= STEP_SIZE ||
       fabs( _position.y() - _destination.y() ) >= STEP_SIZE )
  {
    if ( _lastVisibility != _visible )
    {
      sendOutput( msg.time(), _engaged, _visible );
      _lastVisibility = _visible;
    }
    // Figure out position of next step
    Point offset = _destination - _position;
    if ( is_not_equal( offset.x(), static_cast< Value >( 0 ) ) &&
         is_not_equal( offset.y(), static_cast< Value >( 0 ) ) )
    {
      Point::coord_t abs_x = fabs( offset.x() );
      Point::coord_t abs_y = fabs( offset.y() );

      Point::coord_t slope = abs_y / abs_x;
      if ( abs_x > abs_y )
      {
        if ( offset.y() < 0 )
          offset.y( slope * NEGATIVE_STEP_SIZE );
        else
          offset.y( slope * STEP_SIZE );
        if ( offset.x() < 0 )
          offset.x( NEGATIVE_STEP_SIZE );
        else
          offset.x( STEP_SIZE );
      }
      else
      {
        if ( offset.x() < 0 )
          offset.x( (1/slope) * NEGATIVE_STEP_SIZE );
        else
          offset.x( (1/slope) * STEP_SIZE );
        if ( offset.y() < 0 )
          offset.y( NEGATIVE_STEP_SIZE );
        else
          offset.y( STEP_SIZE );
      }
    }
    else if ( is_not_equal( offset.x(), static_cast< Value >( 0 ) ) )
    {
      if ( offset.x() < 0 )
        offset.x( NEGATIVE_STEP_SIZE );
      else
        offset.x( STEP_SIZE );
    }
    else if ( is_not_equal( offset.y(), static_cast< Value >( 0 ) ) )
    {
      if ( offset.y() < 0 )
        offset.y( NEGATIVE_STEP_SIZE );
      else
        offset.y( STEP_SIZE );
    }
    _position += offset;
    sendOutput( msg.time(), _nextPosition, _position.x() );
    sendOutput( msg.time(), _nextPosition, _position.y() );
  }
  else
  {
    if ( _outputSuccess )
    {
      sendOutput( msg.time(), _out, _success );
      _outputSuccess = false;
    }
    sendOutput( msg.time(), _ready, count++ );
  }
	return *this;
}

std::string Planner::className() const
{
	return "Planner";
}
