#ifndef PLANNER_H
#define PLANNER_H
/**
 * @file planner.h
 * @author Keith Holman
 * @date Created: Tuesday, October 14, 2008
 *
 * @brief Defines the atomic model for the planner
 */

#include "atomic.h"     // class Atomic
#include "serial.h"
#include "point.h"

#include <queue>

/**
 * @author Keith Holman
 * @date Created: Tuesday, October 14, 2008
 *
 * @brief Atomic model for the planner
 */
class Planner : public Atomic
{
public:
  /**
   * @brief Constructor
   * @param name      name of the atomic model
   */
  Planner( const std::string& name = "Planner" );
  /**
   * @brief returns the name of the class of this atomic model
   * @return the name of the class of this atomic model
   */
  virtual std::string className() const;

protected:
  /**
   * @brief Initialization function
   * @return a reference to @c this object
   */
	Model &initFunction();
  /**
   * @brief External transition function
   * @param msg   transition message object
   * @return a reference to @c this object
   */
	Model &externalFunction( const ExternalMessage &msg );
  /**
   * @brief Internal transition function
   * @param msg   transition message object
   * @return a reference to @c this object
   */
	Model &internalFunction( const InternalMessage &msg );
  /**
   * @brief Output function
   * @param msg   transition message object
   * @return a reference to @c this object
   */
	Model &outputFunction( const InternalMessage &msg );

private:
  enum InternalIndexes
  {
    VISIBLE,
    X_COORD,
    Y_COORD
  };

  /**
   * @brief Copy constructor
   * @param obj       object to copy
   */
  Planner( const Planner &obj );
  /**
   * @brief Assignment operator
   * @param obj       object to copy
   * @return a reference to @c this object
   */
  Planner& operator=( const Planner &obj );

  // Ports
  /** @brief Input port */
  const Port &_in;
  /** @brief Input port */
  const Port &_done;
  /** @brief Output port */
	Port &_out;
  /** @brief Output port */
  Port &_ready;
  /** @brief Output port */
	Port &_engaged;
  /** @brief Output port */
	Port &_nextPosition;

  /** @brief Object that stores data recieved on the input */
  Serial< Value, 3 > _inputData;

  /** @brief Whether movement caused an error */
  bool _success;
  /** @brief Update the success of the system */
  bool _outputSuccess;
  /** @brief Whether the chalk is visible */
  bool _visible;
  /** @brief Value of the last visibility sent */
  bool _lastVisibility;
  /** @brief Current position */
  Point _position;
  /** @brief Destination position */
  Point _destination;

  /** @brief Maximum amount to step in one direction */
  static Value STEP_SIZE;
  /** @brief Negative value of @c STEP_SIZE */
  static Value NEGATIVE_STEP_SIZE;
};

#endif   // ifndef PLANNER_H
