#ifndef TRANSLATOR_H
#define TRANSLATOR_H
/**
 * @file translator.h
 * @author Keith Holman
 * @date Created: Tuesday, October 14, 2008
 *
 * @brief Defines the atomic model for the translator
 */

#include "atomic.h"     // class Atomic
#include "point.h"
#include "serial.h"
#include <queue>

/**
 * @author Keith Holman
 * @date Created: Tuesday, October 14, 2008
 *
 * @brief Enumeration containing numeric constants used by the system
 */
enum NumericConstants
{
  /** @brief Number of motors within the robot */
  NUMBER_OF_MOTORS = 2
};


/**
 * @author Keith Holman
 * @author Jeremy Kuzub
 * @date Created: Tuesday, October 14, 2008
 *
 * @brief Atomic model for the translator
 */
class Translator : public Atomic
{
public:
  /**
   * @brief Constructor
   * @param name      name of the atomic model
   */
  Translator( const std::string& name = "Translator" );
  /**
   * @brief returns the name of the class of this atomic model
   * @return the name of the class of this atomic model
   */
  virtual std::string className() const;

protected:
  /**
   * @brief Initialization function
   * @return a reference to @c this object
   */
	Model &initFunction();
  /**
   * @brief External transition function
   * @param msg   transition message object
   * @return a reference to @c this object
   */
	Model &externalFunction( const ExternalMessage &msg );
  /**
   * @brief Internal transition function
   * @param msg   transition message object
   * @return a reference to @c this object
   */
	Model &internalFunction( const InternalMessage &msg );
  /**
   * @brief Output function
   * @param msg   transition message object
   * @return a reference to @c this object
   */
	Model &outputFunction( const InternalMessage &msg );

private:
  /**
   * @brief Copy constructor
   * @param obj       object to copy
   */
  Translator( const Translator &obj );
  /**
   * @brief Assignment operator
   * @param obj       object to copy
   * @return a reference to @c this object
   */
  Translator& operator=( const Translator &obj );

  // Ports
  /** @brief Input port */
  const Port &_in;
  /** @brief Input port */
  const Port &_done;
  /** @brief Output port */
	Port &_out;
  /** @brief Output port */
	Port &_motorOut1;
  /** @brief Output port */
  Port &_motorOut2;


	/** @brief Array containing target lengths for each of the motors */
  Value _targetLengths[NUMBER_OF_MOTORS];

  /** @brief Whether the requested translation has been completed */
  bool _complete;
  /** @brief Whether movement caused an error */
  bool _success;
  /** @brief Number of done inputs recieved */
  unsigned int _doneCount;

  /** @brief Object that stores data recieved on the input */
  Serial< Value, NUMBER_OF_MOTORS > _inputData;

  /** @brief Anger points where the motors are connected to the outside */
  Point _anchorPoints[NUMBER_OF_MOTORS];
};

#endif   // ifndef TRANSLATOR_H
