/*******************************************************************
*
*  DESCRIPTION: class Coordinator
*
* Modification History:
*
* yyyymmmdd     Name            Description
* =========     ========        ===========
* 1998Jun27     Barylko & Beyoglonian     Initial Creation
* 1999Jun04     D. Rodriguez    V2.0
* 2004Apr04     Henry Yu        V3.0: Flat Coordinator with Parallel DEVS
* 2004Oct20     Henry Yu        V4.2: Bug fix in Done Msg Handler
*
*
*******************************************************************/

/** include files **/
#include "coordin.h"       // header
#include "msgadm.h"        // SingleMsgAdmin 
#include "coupled.h"       // class Coupled
#include "procadm.h"
#include "debug.h"	//DBG

/*******************************************************************
* Function Name: Coordinator
********************************************************************/
Coordinator::Coordinator( Coupled *coupled )
: Processor( coupled ), 
doneCount_m( 0 ), 
receivedX_m( 1 ), 
waitingMode_m( noWaiting_c )
{}

/*******************************************************************
* Function Name: ~Coordinator
********************************************************************/
Coordinator::~Coordinator()
{
  immiChildren_m.clear();
  syncSet_m.clear();
  collectWaitForDoneQ_m.clear();
  //outputWaitForDoneQ_m.clear();
}

/*******************************************************************
* Function Name: receive
* Description: forward the init message to his childrens
********************************************************************/
Processor &Coordinator::receive( const InitMessage &msg )
{
  Coupled &coupled( static_cast<Coupled &>( model() ) );

  immiChildren_m.clear();
  syncSet_m.clear();
  collectWaitForDoneQ_m.clear();
  //outputWaitForDoneQ_m.clear();
  waitingMode_m = waitingUponRecvInitMsg_c;
  doneCount_m = coupled.children().size();
 
  lastChange(msg.time());

  //Set next change to Inf. It will be updated when all the 
  //(done, t)'s are received from all the children.
  nextChange(Time::Inf);

  InitMessage init( msg.time(), this->Processor::id() );
  Coupled::ModelList::const_iterator cursor;
  for( cursor = coupled.children().begin(); 
       cursor != coupled.children().end(); cursor++ )
  {
    immiChildren_m.push_back(*cursor);
    SingleMsgAdm::Instance().send( init , *cursor );
  }//for

  return *this;
}

/*******************************************************************
* Function Name: receive CollectMessage (a)
* Description: forward the @ message to inminent children
********************************************************************/
Processor &Coordinator::receive( const CollectMessage &msg )
{
  if (absoluteNext() == msg.time())
  {
    lastChange( msg.time() );

    CollectMessage collectMsg( msg.time(), this->Processor::id() );

    MASSERT(waitingMode_m == noWaiting_c);
    if(!collectWaitForDoneQ_m.empty()) 
    {
      InvalidMessageException e ;
      e.addText( std::string( "collectWaitForDoneQ_m not empty" ));
      MTHROW( e ) ;
    }//if

    ImminentChildren::const_iterator cursor;
    for( cursor = immiChildren().begin();
         cursor != immiChildren().end(); cursor++ )
    {
      collectWaitForDoneQ_m.insert(*cursor);
      
      SingleMsgAdm::Instance().send( collectMsg , *cursor );

      //Add inminent child to synchronize set if it is not there.
      SyncAtomicsSet::const_iterator p = syncSet_m.find(*cursor);
      if (p == syncSet_m.end())
      {
        syncSet_m.insert(p, *cursor); //add *cursor before p
      }//if
    }//for

    //printSyncSet_v();

    //wait until (done, t)'s are received from all imminent children.
    //Then send (done, t) to the parent coordinator
    waitingMode_m = waitingUponRecvCollectMsg_c;
  }
  else
  {
    InvalidMessageException e ;
    e.addText( std::string( "Wrong CollectMsg timing\n" ));
    e.addText(nextChange().asString());
    e.addText(msg.asString());
    MTHROW( e ) ;
  }//if-else
  return *this;
}//receive CollectMessage

/*******************************************************************
* Function Name: receive q message
* Description: add q to msg bag
********************************************************************/
Processor &Coordinator::receive( const ExternalMessage &msg )
{
  //msg.port().print(std::cout);
  externalMsgs_m.add (static_cast<ExternalMessage *>(msg.clone()));

  return *this ;
}

/*******************************************************************
* Function Name: receive * msg
* Description: sends the * message to the inminent child
********************************************************************/
Processor &Coordinator::receive( const InternalMessage &msg )
{
  MASSERT( (msg.time() >= lastChange()) && (msg.time() <= absoluteNext()) );
  Coupled &coupled( static_cast<Coupled &>( model() ) );
  int value = (int)msg.value();
if (value != 0) // structural change 
{
 std::cout<<"Coordinator receive an internal message, message value is non-zero"<<std::endl;
 
     
  //   if(MainSimulator::Instance().IsStrucComponent(coupled.description())) 
  //	strucset.insert(coupled.executive());
      Coupled::ModelList::const_iterator it = coupled.children().begin();
      for(; it != coupled.children().end(); it++)
	{
            Processor &proc = SingleProcessorAdmin::Instance().processor(*it);
	    std::string mName = proc.model().description();
            if((proc.model().mType() == Coupled::coupledType) && (proc.model().executive() != Processor::InvalidId/*MainSimulator::Instance().IsStrucComponent(mName)*/))
	      {
		
                  StructuredSet::iterator cur = strucset.begin();

		  cur = strucset.find(proc.model().id());
		  if(cur == strucset.end() )        	   

                     strucset.insert(proc.model().id());
		 
	      }
	}
      std::cout<<"strucset size is "<<strucset.size()<<std::endl;
      if(strucset.size() != 0)
	{
      StructuredSet::const_iterator scursor = strucset.begin();
     std::cout<<"strucset size is "<<strucset.size()<<" "<<(*scursor)<<std::endl;

      // coupled.oldchildren();
      //coupled.printoldchildren();
      coupled.setvalue(value);
      for(; scursor != strucset.end(); scursor++)
	{
          InternalMessage internal(msg.time(), this->Processor::id(), value);
	  SingleMsgAdm::Instance().send(internal, *scursor);
        }
      waitingMode_m = waitingUponRecvStrucChangeMsg_c;
      lastChange(msg.time());
      doneCount_m = strucset.size();
	}
      else{
	waitingMode_m = waitingUponRecvSelfStrucChange_c;
        coupled.oldchildren();
        coupled.printoldchildren();
        if(coupled.exective() != Processor::InvalidId) 
	  {
            InternalMessage internal(msg.time(), this->Processor::id(), value);
	    SingleMsgAdm::Instance().send(internal, coupled.executive());
	  }
	   doneCount_m = 1;
      }

}
 else{
  //externalMsgs_m.print_v();	//debug

  //For each message in msg bag, send External msg to all its influences. 
  MessageBag::iterator bi = externalMsgs_m.begin(); //bag cursor
  for( ; bi != externalMsgs_m.end(); bi++ )
  {
    ExternalMessage xMsg(*(( ExternalMessage* )( *bi )) );
    xMsg.procId( this->Processor::id() ) ;
    //xMsg.port().print(std::cout);

    const InfluenceList &influList( xMsg.port().influences() ) ;
    InfluenceList::const_iterator pc( influList.begin() ); //port cursor

    for( ; pc != influList.end(); pc++ )
    {
      xMsg.port( *(*pc) ) ;
      SingleMsgAdm::Instance().send( xMsg, (*pc)->model() ) ;

      //Add procId to synchronize set if it is not there.
      //if (syncSet_m.size() == 0){
      syncSet_m.insert((*pc)->model());
#if 0
      }else{
        SyncAtomicsSet::const_iterator p = syncSet_m.find((*pc)->model());
        if (p == syncSet_m.end())
        {
          syncSet_m.insert(p, (*pc)->model()); //add model before p
        }//if
      }//if-else
#endif
    }//for
  }//for

  //empty msgbag
  externalMsgs_m.eraseAll();

  //printSyncSet_v();

  //send (*, t) to all syncSet members
  SyncAtomicsSet::const_iterator si;  //syncSet iterator
  for (si = syncSet_m.begin(); si != syncSet_m.end(); si++)
  {
    InternalMessage internal( msg.time(), this->Processor::id() );
    SingleMsgAdm::Instance().send( internal, *si );
  }//for

  lastChange(msg.time());

  //wait until (done, t)'s are received from all children in synchSet.
  //Then send (done, t) to the parent coordinator and clear syncSet_m
  waitingMode_m = waitingUponRecvInternalMsg_c;
  doneCount_m = syncSet_m.size();
}

  //clear the synchronize set
  //Don't do it here. Do it in receive(Done).
  //syncSet_m.clear();

  return *this ;
}//receive InternalMessage

/*******************************************************************
* Function Name: receive Y msg
* Description: translates the output event to a X messages for the 
*              influenced children or to a Y message for his father
********************************************************************/
Processor &Coordinator::receive( const OutputMessage &msg )
{
  //We decide not to use the outputWaitForDoneQ_m.
  //MASSERT(outputWaitForDoneQ_m.empty());

  //This assertion is not true, because not only can collectMsg 
  //receiver send Y msgs, OutputMsg receiver can also send Y msgs.
  //MASSERT(waitingMode_m == waitingUponRecvCollectMsg_c);

  Coupled &coupled( static_cast< Coupled & >( model() ) );

  const InfluenceList &influList( msg.port().influences() );

  InfluenceList::const_iterator cursor( influList.begin() );

  OutputMessage outMsg( msg );
  outMsg.procId( this->Processor::id() );

  ExternalMessage extMsg;
  extMsg.time( msg.time() );
  extMsg.procId( Processor::id() );
  extMsg.value( msg.value() );

  for( ; cursor != influList.end(); cursor++ )
  {
    // not found in the output list => send it as an external message
    if( coupled.outputPorts().find( (*cursor)->id() ) == coupled.outputPorts().end() )
    {
      //outputWaitForDoneQ_m.insert((*cursor)->model());
      extMsg.port( * (*cursor) );
      SingleMsgAdm::Instance().send( extMsg, (*cursor)->model() );

      //Add child to synchronize set if it is not there.
      SyncAtomicsSet::const_iterator p = syncSet_m.find((*cursor)->model());
      if (p == syncSet_m.end())
      {
        syncSet_m.insert(p, (*cursor)->model()); //add ProcId before p
      }//if

      //printSyncSet_v();

      //The papers says we should wait until all (done t)'s are received from 
      //all children. Then send (Y t) to the parent coordinator.
      //However, we decide not to do this, as it is not really necsissary
      //to wait for all Done's; so, comment the following line out.
      //waitingMode_m = waitingUponRecvOutputMsg_c;
    }//if
    else
    {
      outMsg.port( * (*cursor) );
      SingleMsgAdm::Instance().send( outMsg, coupled.parentId() );
    }//else
  }//for
  return *this;
}//receive OutputMessage

/*******************************************************************
* Function Name: receive D msg
* Description: recalculates the inminent child and sends it to his 
*              father inside a done message
********************************************************************/
Processor &Coordinator::receive( const DoneMessage &msg )
{
  Coupled &coupled( static_cast< Coupled & >( model() ) );

  /*
  MASSERTMSG( (waitingMode_m == waitingUponRecvCollectMsg_c)||
              (waitingMode_m == waitingUponRecvInternalMsg_c), 
	      "Unexpected Done message!" );
  */

  switch(waitingMode_m)
  {
    case waitingUponRecvInitMsg_c:
    {
      if( ! --doneCount_m )
      {
	waitingMode_m = noWaiting_c;

        lastChange( msg.time() );

        recalcImminentChildren();

        if( imminentChild() == Processor::InvalidId )
        {
          nextChange( Time::Inf );
        }else {
          Processor &proc = SingleProcessorAdmin::Instance().
                                 processor(*(immiChildren().begin()) );
          nextChange( proc.model().absoluteNext() - msg.time() );
        }//if-else

        DoneMessage doneMsg(msg.time(), this->Processor::id(), 
	                                      coupled.nextChange());
        SingleMsgAdm::Instance().send( doneMsg, coupled.parentId() );
      }//if
      break;
    }//case waitingUponRecvInitMsg_c

    case waitingUponRecvCollectMsg_c:
    {
      WaitForDoneQueue::const_iterator p; 
      p = collectWaitForDoneQ_m.find(msg.procId());
      if (p == collectWaitForDoneQ_m.end())
      {
        InvalidMessageException e ;
        e.addText( std::string( "procId not in collectWaitForDoneQ_m!" ));
        MTHROW( e ) ;
      }else{
        collectWaitForDoneQ_m.erase(p); 
      }//if-else

      if (collectWaitForDoneQ_m.size() == 0)
      {
        //We have now receive all the Ds from imminent children.
	//So, send Done to the parent.

	waitingMode_m = noWaiting_c;

	//This Done message should NOT trigger Root to advance its time,
	//because time advances should only occur during (*, t)
	//messages.
        //DoneMessage doneMsg(msg.time(), this->Processor::id(), 
	//					coupled.nextChange());
        DoneMessage doneMsg(msg.time(), this->Processor::id(), Time::Zero);
        SingleMsgAdm::Instance().send( doneMsg, coupled.parentId() );
      }//if

      break;
    }//waitingUponRecvCollectMsg_c case
    
    case waitingUponRecvInternalMsg_c:
    {
      //For Debug Only: check to see if child is in synchronize set.
      SyncAtomicsSet::const_iterator p = syncSet_m.find(msg.procId());
      if (p == syncSet_m.end())
      {
        InvalidMessageException e ;
        e.addText( std::string( "procId not in syncSet_m!" ));
        MTHROW( e ) ;
      }else{
        syncSet_m.erase(p);
      }//if-else
      //structural request may be attached with the internal message.
      int value = (int)msg.value();
      if(value > 0) screq[msg.procId()] = value;

      if( ! --doneCount_m )
      {
        //We have received all the Done msgs from children 
        lastChange( msg.time() );

        recalcImminentChildren();

        if( *(immiChildren().begin()) == Processor::InvalidId )
	{
          nextChange( Time::Inf );
        }else {
          Processor &proc = SingleProcessorAdmin::Instance().
                                 processor(*(immiChildren().begin()) );
          nextChange( proc.model().absoluteNext() - msg.time() );
        }//if-else
          if(!screq.empty())
           {
	     SCReq::iterator it = screq.begin();
             std::cout<<"screq is not empty the element is "<<(*it).first<<" "<<(*it).second<<std::endl;	     
	    
             ImminentChildren::const_iterator imm_cursor = immiChildren().begin();
	     while (imm_cursor != immiChildren().end())
	       {
	     if ((it->first) == (*imm_cursor))
	       {
            std::cout<<"find out a structure component from imminent children "<<(*imm_cursor)<<std::endl;	
	    std::cout<<"the msg value is "<<it->second<<std::endl;
	    std::cout<<"the sender is "<<coupled.description()<<" receiver is "<<coupled.parentId()<<std::endl;
	     waitingMode_m = noWaiting_c;          
            DoneMessage doneMsg(msg.time(), this->Processor::id(), coupled.nextChange(), it->second);
            SingleMsgAdm::Instance().send(doneMsg, coupled.parentId());
            std::cout<<"the msg value is "<<it->second<<std::endl;
	    Processor &proc = SingleProcessorAdmin::Instance().processor(*imm_cursor);
	    if(proc.model().mType() == Coupled::coupledType)
	      {
	         ModelId mid = proc.model().id();
	         strucset.insert(mid);
	      }
	    
	    break;
	       }
	     else imm_cursor++;

	       }
	    
	     screq.clear();
           }
        //V4.2: bug fix - We cannot clear syncSet_m, because it may
        //contain elements that were inserted by the Output Message
        //Handler during the waitingUponRecvInternalMsg_c state.
        //So, instead of clear the syncSet_m when doneCount_m becomes 0,
        //we erase the element from sysncSet_m one at a time every time we
        //receive the done msg.
        //syncSet_m.clear();
	  else{

	waitingMode_m = noWaiting_c;

        DoneMessage doneMsg(msg.time(),this->Processor::id(), 
	                                      coupled.nextChange());
        SingleMsgAdm::Instance().send( doneMsg, coupled.parentId() );
	  }

      }//if
      break;
    }//waitingUponRecvInternalMsg_c case
case waitingUponRecvStrucChangeMsg_c:
    {
      if(!--doneCount_m)
	{
	  //bool IsStruc = MainSimulator::Instance().IsStrucComponent(coupled.description());
         if(coupled.exective() != Processor::InvalidId)  
           {
	     std::cout<<"the current coupled model "<<coupled.description()<<" is structure component"<<std::endl;
              waitingMode_m = waitingUponRecvSelfStrucChange_c;
              coupled.oldchildren();
              coupled.printoldchildren();
	      int value = coupled.getvalue();
	      doneCount_m = 1;
              InternalMessage internal(msg.time(), this->Processor::id(), value);
              SingleMsgAdm::Instance().send(internal, coupled.executive());
           }
	 else{
              recalcImminentChildren();
              
              if( *(immiChildren().begin()) == Processor::InvalidId )
        	{
                  nextChange( Time::Inf );
                }else {
                  Processor &proc = SingleProcessorAdmin::Instance().
                                 processor(*(immiChildren().begin()) );

	          nextChange(proc.model().absoluteNext() - msg.time());        
	       }
	  waitingMode_m = noWaiting_c;
	  DoneMessage doneMsg(msg.time(), this->Processor::id(), coupled.nextChange());
	  SingleMsgAdm::Instance().send(doneMsg, coupled.parentId());
	 }
       
          lastChange(msg.time());
	  
	}
      break;
    }
 case waitingUponRecvSelfStrucChange_c:
    {
      
      Processor &proc = SingleProcessorAdmin::Instance().processor(msg.procId());
      std::cout<<"Coordinator receive structure done message"<<"sender is "<<msg.procId()<<std::endl;
      coupled.setremmodels();
      if(!coupled.getremmodels().empty())
	{
	  
	  Coupled::ModelList::const_iterator rem_cur(coupled.getremmodels().begin());
          for(; rem_cur != coupled.getremmodels().end(); rem_cur++)
	    {
	      SyncAtomicsSet::const_iterator p = syncSet_m.find(*rem_cur);
	      if(p != syncSet_m.end()) syncSet_m.erase(p);
	    }
	}
      coupled.setnewmodels();
      if((proc.model().mType() == Model::revatomicType) && (!coupled.getnewmodels().empty()))
	{
	 
          StartMessage startmsg(msg.time(), this->Processor::id());
	  Coupled::ModelList::const_iterator mcursor(coupled.getnewmodels().begin());
	  waitingMode_m = waitingUponRecvStartMsg_c;
	  for(; mcursor !=coupled.getnewmodels().end(); mcursor++)
	    {
	  SingleMsgAdm::Instance().send(startmsg, *mcursor);
	    }
	  doneCount_m = coupled.getnewmodels().size();
	}
	 else/* if(proc.model().mType() == Model::coupledType)*/
	{
	  if((proc.model().mType() == Model::revatomicType) && (coupled.getnewmodels().empty()))   
	    std::cout<<"no new models, no need to send start messages !! "<<doneCount_m<<std::endl;  
	 
	  if(!--doneCount_m)
	    {
	      recalcImminentChildren();
              
              if( *(immiChildren().begin()) == Processor::InvalidId )
        	{
                  nextChange( Time::Inf );
                }else {
                  Processor &proc = SingleProcessorAdmin::Instance().
                                 processor(*(immiChildren().begin()) );

	          nextChange(proc.model().absoluteNext() - msg.time());        
	       }
          lastChange(msg.time());
	  waitingMode_m = noWaiting_c;
	  DoneMessage doneMsg(msg.time(), this->Processor::id(), coupled.nextChange());
	  SingleMsgAdm::Instance().send(doneMsg, coupled.parentId());
	    }
	}
      break;
    }
  case waitingUponRecvStartMsg_c:
    {
      if(!--doneCount_m)
	{
         recalcImminentChildren();
         if( *(immiChildren().begin()) == Processor::InvalidId )
           {
              nextChange( Time::Inf );
           }else {
              Processor &proc = SingleProcessorAdmin::Instance().
                                 processor(*(immiChildren().begin()) );

	      nextChange(proc.model().absoluteNext() - msg.time());        
	   }
      waitingMode_m = noWaiting_c;
      DoneMessage doneMsg(msg.time(), this->Processor::id(), coupled.nextChange());
      SingleMsgAdm::Instance().send(doneMsg,coupled.parentId());
	}
      break;
    }
    default:
    {
      InvalidMessageException e ;
      e.addText( std::string( "Unexpected Done message!" ));
      MTHROW( e ) ;
    }//default case
  }//switch

  return *this;
}
/*******************************************************************
* Function Name: receive St msg
* Description: 
********************************************************************/
Processor &Coordinator::receive(const StartMessage &msg)
{
  Coupled &coupled( static_cast< Coupled & >( model() ) );
  recalcImminentChildren();
  if( *(immiChildren().begin()) == Processor::InvalidId )
           {
              nextChange( Time::Inf );
           }else {
              Processor &proc = SingleProcessorAdmin::Instance().
                                 processor(*(immiChildren().begin()) );

	      nextChange(proc.model().absoluteNext() - msg.time());        
	   }
	   DoneMessage doneMsg(msg.time(), this->Processor::id(), coupled.nextChange());
  return *this;
}
/*******************************************************************
* Function Name: id
********************************************************************/
Processor &Coordinator::id( const ProcId &pid )
{
  this->Processor::id( pid );

  MASSERT( pmodel() );
  pmodel()->id( pid );

  return *this;
}

/*******************************************************************
* Function Name: recalcImminentChildren
********************************************************************/
Coordinator &Coordinator::recalcImminentChildren()
{
  Coupled &coupled( static_cast< Coupled & >( model() ) );

  Coupled::ModelList::const_iterator cursor( coupled.children().begin() ) ;

  Time min( Time::Inf );

  for( ; cursor != coupled.children().end() ; cursor++ )
  {
    Model &model( SingleProcessorAdmin::Instance().processor( *(cursor) ).model() ) ;
    if( min > model.absoluteNext() )
    //HENRY compile fix disabled
    //if( min >= model.absoluteNext() )
    {
      min = model.absoluteNext();
    }
  }//for
 
  immiChildren_m.clear();

  if (min == Time::Inf)
  {
    immiChildren_m.push_back(Processor::InvalidId);
    return *this;
  }//if

  for(cursor = coupled.children().begin();
      cursor != coupled.children().end() ; cursor++ )
  {
    Model &model(SingleProcessorAdmin::Instance().processor(*(cursor)).model());
    if( min == model.absoluteNext() )
    {
      std::string str = "add " + model.asString();
      str += " ["; 
      str += model.absoluteNext().asString();
      str += "] to "; 
      str += coupled.asString(); 
      str += "'s immiChildren_m"; 
      DBG(str);
      immiChildren_m.push_back(*cursor);
    }
  }//for 

  return *this;
}

/*******************************************************************
* Function Name: printSyncSet_v
********************************************************************/
void Coordinator::printSyncSet_v()
{
  std::cout << "\nCoordinator::printSyncSet_v():\n";
  SyncAtomicsSet::const_iterator si;  //syncSet iterator
  for (si = syncSet_m.begin(); si != syncSet_m.end(); si++)
  {
    std::cout << SingleProcessorAdmin::Instance().processor(*si).
                			model().asString() << std::endl;
  }//for

  return;
}//printSyncSet_v
void Coordinator::printStrucSet_v()
{
  return;
}
