/*******************************************************************
*
*  DESCRIPTION: Draws the content of the log 
*
*  AUTHOR:    Amir Barylko & Jorge Beyoglonian 
*  Version 2: Daniel Rodriguez
*
*  EMAIL: mailto://amir@dc.uba.ar
*         mailto://jbeyoglo@dc.uba.ar
*         mailto://drodrigu@dc.uba.ar
*
*  DATE: 27/06/1998
*  DATE: 28/04/1999 (v2)
*
*******************************************************************/

// ** include files **//
#include <strstream>
#include <fstream>
#include <ini.h>     // Class Ini
#include <cdTime.h>    // Class Time
#include <strutil.h> // str2Int y str2Real
#include <cellstate.h>

// ** main ** //

void printState( const CellState &state, const Time &time ) 
{
	std::cout << "Time: " << time.asString() << std::endl ;
	state.print(std::cout, '?');
}

void showHelp()
{
	std::cout << "drawlog -[?hmtclwp0]\n\n";
	std::cout << "where:\n";
	std::cout << "\t?\tShow this message\n";
	std::cout << "\th\tShow this message\n";
	std::cout << "\tm\tSpecify file containig the model (.ma)\n";
	std::cout << "\tt\tInitial time\n";
	std::cout << "\tc\tSpecify the coupled model to draw\n";
	std::cout << "\tl\tLog file containing the output generated by SIMU\n";
	std::cout << "\tw\tWidth (in characters) used to represent numeric values\n";
	std::cout << "\tp\tPrecision used to represent numeric values (in characters)\n";
	std::cout << "\t0\tDon't print the zero value\n";
        std::cout << "\tf\tOnly cell values on a specified slice in 3D models\n";
	exit(1);
}

bool parseLine( const char *l, Time &current, const std::string &modelName, CellPosition &posi, Real &value )
{
	std::istrstream line( l ) ;

	std::istream_iterator<std::string> cursor( line ) ;
	std::string tokens[ 12 ];
	register long pos( 0 ) ;

	while( cursor != std::istream_iterator<std::string>() && pos < 12 )
		tokens[ pos++ ] = *cursor++ ;

	if( pos == 12 && tokens[0] == "Mensaje" && tokens[1] == "Y" && tokens[5].find( modelName ) == 0 && tokens[11].find( modelName ) == 0 )
	{
		// the time of the message
		current = tokens[ 3 ] ; 

		CellPosition pos( tokens[5] ) ;
		posi = pos ;
		value = Real( str2Real(tokens[ 9 ]) ) ;
		return true ;
	}
	return false ;
}

int splitLine( char *line, char *pos, char *value)
{
        char    *posi = pos, *val = value;

        // Primero leo POS
        while ( *line != 0 && *line != '=' )
        {
                if (*line != ' '){
                        *posi = *line;
                        posi++;
                }
                line++;
        }
        if (*line == 0)
                return 0;

        *posi = 0;
        line++;

        // Ahora leo el value
        while ( *line != 0 )
        {
                if (*line != ' '){
                        *val = *line;
                        val++;
                }
                line++;
        }
        *val = 0;

        if (pos[0] == '(' && strlen(value) > 0)
                return 1;

        return 0;
}


//////////////////////////////////////////////////////////////////////////////
// MAIN - DRAWLOG
//////////////////////////////////////////////////////////////////////////////
int main( int argc, char *argv[] )
{
	try
	{
		Time initial( Time::Zero );
		std::string modelName, iniName, logName("-"), strWidth(""), strPrec(""), strPlane("");

		// parameter parsing

		while( --argc )
			if( *argv[ argc ] == '-' )
				switch( argv[ argc ][ 1 ] )
				{
					case 'm': /* file .ma */
						iniName = lowerCase(argv[ argc ] + 2);
						break;

					case 't': /* intital time */ 
						initial = argv[ argc ] + 2 ;
						break;

					case 'c': /* coupled */
						modelName = lowerCase(argv[ argc ] + 2);
						break;

					case 'l': /* log filename */
						logName = argv[ argc ] + 2 ;
						break;

					case 'w': /* Set width */
						strWidth = argv [ argc ] + 2;
						Impresion::Default.Width( str2Int(strWidth) );
						break;

					case 'p': /* Set precision */
						strPrec = argv [ argc ] + 2;
						Impresion::Default.Precision ( str2Int(strPrec) );
						break;

					case '0': /* Don't print zero */
						Impresion::Default.PrintZero(false);
						break;

					case '?':
					case 'h':
						showHelp();
						break;

					case 'f': /* Only cell values on 3D models */
						Impresion::Default.FlatLog(true);
						strPlane = argv [ argc ] + 2;
						Impresion::Default.FlatLogPlane( str2Int(strPlane) );
						break;
 
					default:
						std::cout << "Warning... invalid parameter " << argv[ argc ] << "!" << std::endl ;
						showHelp();
				}
			else
				std::cout << "Warning... invalid parameter " << argv[ argc ] << "!"	<< std::endl ;

		// parameter validation

		if( iniName == "" || modelName == "" )
		{
			std::cout << "Usage: " << argv[ 0 ] << " -mfile.ma -cCoupledCellName [ -tInitialtime ] \n\t[ -lmessage.log ] [ -wWidth ] [ -pPrecision ] [ -0 ]" << std::endl;
			return 1 ;
		}

		Ini iniFile ;

		iniFile.parse( iniName, false ) ;

		// dimension
		nTupla			nt;
		register unsigned	cols = 0, rows = 0;
		if (iniFile.exists( modelName, "width" ))
		{
			cols = str2Int( iniFile.definition( modelName, "width"  ).front() );
			rows = str2Int( iniFile.definition( modelName, "height" ).front() );

			nt.add(rows, cols);
		}
		else if (iniFile.exists( modelName, "dim" ))
		{
			Ini::IdList	dimension = iniFile.definition( modelName, "dim" );
			CellPosition	cp( iniFile.join(dimension) );

			nt = cp;

			cols = nt.get(DIM_WIDTH);
			rows = nt.get(DIM_HEIGHT);
		}

		if ( nt.contains(0) )
			MASSERT_ERR( "Attemp to draw a model where a component of its dimension is 0" );

		CellState state( nt ) ;

		//////////////////////////////////////////////////////////////
		// default initial value
		std::string initialValue( iniFile.definition( modelName, "initialvalue" ).front() ) ;

		Real value ( str2Real( initialValue ) );

		CellPosition	counter( nt.dimension(), 0);
		register bool	overflow = false;

		while (!overflow)
		{
			state[ counter ] = value;
			overflow = counter.next( nt );
		}
		/////////////////////////////////////////////////////////////

		/////////////////////////////////////////////////////////////
		// loading the initial state
		/////////////////////////////////////////////////////////////
		if( iniFile.exists( modelName, "initialrow" ) )		// The dim = 2
		{
			const Ini::IdList &values( iniFile.definition( modelName, "initialrow" ) ) ;
			register unsigned row = 0;
			Ini::IdList::const_iterator cursor = values.begin();

			while (cursor != values.end() )
			{
				// el primer valor es la fila
				row = str2Int( (*cursor) ) ;
				MASSERTMSG( row <= rows-1, "The number of row for initialRowValue is out of range. It's " + int2Str(row) + " and must be in [ 0, " + int2Str( rows - 1 ) + "]" );

				cursor++ ;

				// Los siguientes elementos son la descripcion de la fila
				register unsigned col = 0;

				while ( col < cols )
				{
					MASSERTMSG( cursor != values.end(), "Insuficient data for initialRowValue. Last row with " + int2Str(col) + " elements. Note: May be a middle row definition with std::less elements.");
					std::string rowVal( (*cursor) ) ;

					nTupla	nt3;
					nt3.add(row, col);
					state[ nt3 ] = str2Real( rowVal ) ;
					col++ ;
					cursor++ ;
				}
			}
		}

		if( iniFile.exists( modelName, "initialRowValue" ) )	// The dim = 2
		{
			const Ini::IdList &rowsList( iniFile.definition(modelName, "initialRowValue" ) ) ;

			register unsigned row = 0;
			Real val;

			for ( Ini::IdList::const_iterator cursor = rowsList.begin(); cursor != rowsList.end(); cursor++ )
			{
				// the first value is the row number
				row = str2Int( (*cursor) ) ;
				MASSERTMSG( row < rows, "The number of row for initialRowValue is out of range. It's " + int2Str(row) + " and must be in [ 0, " + int2Str( rows-1 ) + "]" ) ;

				cursor++ ;

				MASSERTMSG( cursor != rowsList.end(), "Invalid initial row value for initialRowValue (must be a pair rowNumber rowValues)!" );

				// the second is the description of the row
				std::string rowVal( *cursor ) ;			

				MASSERTMSG( rowVal.size() == cols, "The size of the rows for the initial values of the CoupledCell must be equal to the width value !" );

				register unsigned col = 0;

				for (std::string::iterator rowCurs = rowVal.begin(); rowCurs != rowVal.end(); rowCurs++ )
				{
					if (*rowCurs >= '0' && *rowCurs <= '9')
						val.value( *rowCurs - '0');
					else
						val = Real::tundef;

					nTupla nt4;
					nt4.add(row, col);
					state[ nt4 ] = val;
					col++;
				}
			}
		}

		if( iniFile.exists( modelName, "initialCellsValue" ) )
		{
			std::string	fileName( iniFile.definition( modelName, "initialCellsValue" ).front() );
			FILE	*fileIn;
			char	line[250], pos[200], value[50];

			fileIn = fopen( trimSpaces(fileName), "r" );
			MASSERTMSG( fileIn != NULL, "Can't open the file '" + fileName + "' defined by the initialCellsValue clause");

			while (!feof(fileIn))
			{
				fgets(line, 255, fileIn);
				if (line != NULL & splitLine(line, pos, value)){
					CellPosition	cp2(pos);
					state[ cp2 ] = str2Real(value);
				}
			}
			fclose( fileIn );
		}

		if( iniFile.exists( modelName, "initialMapValue" ) )
		{
			std::string	fileName( iniFile.definition( modelName, "initialMapValue" ).front() );
			FILE	*fileIn;
			char	line[250];

			fileIn = fopen( trimSpaces(fileName), "r" );
			MASSERTMSG( fileIn != NULL, "Can't open the file '" + fileName + "' defined by the initialMapValue clause");

			CellPosition    counter( nt.dimension(), 0 );
			register bool   overflow = false;

			while (!overflow)
			{
				MASSERTMSG( !feof( fileIn ) && fgets(line, 255, fileIn), "Insuficient data in file specified with InitialMapValue");

				state[ counter ] = str2Real(line);

				overflow = counter.next( nt );
			}
			fclose( fileIn );
		}

		//////////////////////////////////////////////////////////////
		// Now, the initial state is already loaded.
		//////////////////////////////////////////////////////////////

		Time currentTime( Time::Zero ), lastTime( initial ) ;
		istream *log ;

		if( logName == "-" )
			log = &std::cin ;
		else
			log = new std::ifstream( logName.c_str() ) ;

		char line[ 2048 ] = { '\0'} ;
		register bool val = false ;
		register int lineCount(1) ;
		CellPosition	nt5;

		do
		{
			// parsing the log
			while( !log->eof() && currentTime <= lastTime )
			{
				if( val )
					state[ nt5 ] = value ;

				lineCount++ ;
				log->getline( line, 2048 ) ;

				val = parseLine( line, currentTime, modelName, nt5, value ) ;
			}

 			if (!Impresion::Default.FlatLog())
				std::cout << "Line : " << lineCount << " - " ;

			printState( state, lastTime ) ;

			lastTime = currentTime ;

		} while( log->good() && !log->eof() ) ;

		if( logName != "-" )
			delete log ;

	} catch( MException &e )
	{
		e.print(std::cerr);
	} catch( ... )
	{
		std::cerr << "Unknown exception! " << std::endl ;
	}
}
