/*******************************************************************
*
*  DESCRIPTION: class FlatDEVSCoordinator
*
* Modification History:
*
* yyyymmmdd	Name		Description
* =========	========	===========
* 2003Jan19	Henry Yu	Initial Creation  
* 2004Apr28	Henry Yu	V3.0: Flat Coordinator with Parallel DEVS
* 2004Oct20	Henry Yu	V4.2: Bug fix in Done Msg Handler
*
*******************************************************************/


/** include files **/
#include "flatdcoo.h"  // header
#include "atomic.h"   // class Atomic
#include "revatomic.h"
#include "coupled.h"  // class Coupled
#include "flatcoup.h"  // class FlatCoupledCell
#include "coupcell.h"  // class CoupledCell
#include "procadm.h"  // class ProcessorAdmin
#include "message.h"  // class Message
#include "mainsimu.h"  //class MainSimulator
#include "msgadm.h"  // class SingleMsgAdm
#include "debug.h"  // DBG

/*******************************************************************/
/** Global Constants **/


/*******************************************************************/

/*******************************************************************
* Function Name: Constructor
* Description  : FlatDEVSCoordinator's constructor, carries out the
* simple construction taks, such as deleting the atomic database
* Receives a coupled model (i.e., the top model) as a parameter
********************************************************************/
FlatDEVSCoordinator::FlatDEVSCoordinator( Coupled * coupled )
  : Processor(coupled),
doneCount_m( 0 ),
waitingMode_m( noWaiting_c )
{

  this->id(coupled->id());

  parentId = model().parentId();

  DBG("FlatDEVSCoordinator created!");

}

#if 0
/*******************************************************************
* Function Name: Constructor
* Description  : (NOT USED) Simple class' constructor, no parameters
********************************************************************/
FlatDEVSCoordinator::FlatDEVSCoordinator( )
{

}
#endif

/*******************************************************************
* Function Name: Destructor
* Description  : Simple class' destructor
********************************************************************/
FlatDEVSCoordinator::~FlatDEVSCoordinator()
{
  immiChildren_m.clear();
  syncSet_m.clear();
  collectWaitForDoneQ_m.clear();
}

/*******************************************************************
* Function Name: receive
* Description:
* 1. Forward the init message to all children
* 2. Calculate the time for next event
* 3. Send Done Message to the Root
********************************************************************/
Processor &FlatDEVSCoordinator::receive( const InitMessage &msg )
{
  DBG("InitMessage Received in FlatDEVSCoordinator");

  immiChildren_m.clear();
  syncSet_m.clear();
  collectWaitForDoneQ_m.clear();

  //For each atomic, execute its init function and 
  //modify the last and next times
  ProcessorAdmin::ProcessorDB::const_iterator cursor;

  for (cursor = SingleProcessorAdmin::Instance().procDB.begin(); 
       cursor != SingleProcessorAdmin::Instance().procDB.end(); 
       cursor++)
  {
    //Model *m_p = (cursor->second)->pmodel();
    //(m_p == 0)? cerr << "null\n" : cerr << m_p->asString() << endl;
    Model &m = (cursor->second)->model();
    //(&m == NULL)? cerr << "null\n" : cerr << m.asString() << endl;
    
    if (&m == NULL) continue;
    //cerr << m.asString() << " " << m.className() << endl;
    if (m.mType() == Model::atomicType)
    {
      //Cast and obtain the atomic to be initialized
      //Now we will refer to the atomic, instead of the model
      Atomic &at( static_cast<Atomic &>(m)); 

      //if (FlatDebug().Active()) cerr << "Associated atomic found: " << 
      //			    at.asString() << endl;

      (cursor->second)->nextChange( Time::Inf ); 
      (cursor->second)->lastChange(msg.time()); 

      // Execute the atomic's init function
      at.initFunction();
    }//if
   else if(m.mType() == Model::revatomicType)
      {
	RevAtomic &revat( static_cast<RevAtomic &>(m));
	(cursor->second)->nextChange( Time::Inf );
	(cursor->second)->lastChange(msg.time());
	revat.InitFunction();
      }
  } // for

  lastChange(msg.time());

  recalcImminentChildren();

  if( imminentChild() == Processor::InvalidId ) {
    nextChange( Time::Inf );
  }else {
    Processor &proc = SingleProcessorAdmin::Instance().
                       processor(*(immiChildren().begin()) );
    nextChange( proc.model().absoluteNext() - msg.time() );
  }//if-else

  DoneMessage done(msg.time(), this->Processor::id(), nextChange());
  SingleMsgAdm::Instance().send(done, model().parentId()); 

  return *this;
}//receive( const InitMessage &msg )

/*******************************************************************
* Function Name: receive CollectMessage (a)
* Description: forward the @ message to inminent children
********************************************************************/
Processor &FlatDEVSCoordinator::receive( const CollectMessage &msg )
{
  if (absoluteNext() == msg.time())
  {
    lastChange( msg.time() );

    CollectMessage collectMsg( msg.time(), this->Processor::id() );

    MASSERT(waitingMode_m == noWaiting_c);
    if(!collectWaitForDoneQ_m.empty())
    {
      InvalidMessageException e ;
      e.addText( std::string( "collectWaitForDoneQ_m not empty" ));
      MTHROW( e ) ;
    }//if

    ImminentChildren::const_iterator cursor;
    for( cursor = immiChildren().begin();
         cursor != immiChildren().end(); cursor++ )
    {
      collectWaitForDoneQ_m.insert(*cursor);

      SingleMsgAdm::Instance().send( collectMsg , *cursor );

      //Add inminent child to synchronize set if it is not there.
      syncSet_m.insert(*cursor);
    }//for

    //printSyncSet_v();

    //wait until (done, t)'s are received from all imminent children.
    //Then send (done, t) to the parent coordinator
    waitingMode_m = waitingUponRecvCollectMsg_c;
  }
  else
  {
    InvalidMessageException e ;
    e.addText( std::string( "Wrong CollectMsg timing\n" ));
    e.addText(nextChange().asString());
    e.addText(msg.asString());
    MTHROW( e ) ;
  }//if-else
  return *this;
}//receive CollectMessage

/*******************************************************************
* Function Name: receive
* Description: receive External Message
********************************************************************/
Processor &FlatDEVSCoordinator::receive( const ExternalMessage &msg )
{
  DBG("ExternalMessage Received in FlatDEVSCoordinator");

  externalMsgs_m.add (static_cast<ExternalMessage *>(msg.clone()));

  return *this;
}//receive( const ExternalMessage &msg )

/*******************************************************************
* Function Name: receive
* Description: receive Internal Message
********************************************************************/
Processor &FlatDEVSCoordinator::receive( const InternalMessage &msg )
{
  DBG("InternalMessage Received in FlatDEVSCoordinator");

  MASSERT( (lastChange() <= msg.time()) && (msg.time() <= absoluteNext()) );
  Coupled &coupled( static_cast<Coupled &>( model() ));
  int value = (int)msg.value();
  if(value != 0)
    {
       std::cout<<"FlatDEVS Coordinator receive an internal message, message value is non-zero"<<std::endl;
       waitingMode_m = waitingUponRecvSelfStrucChange_c;
       coupled.oldchildren();
       coupled.printoldchildren();
       if(coupled.exective() != Processor::InvalidId)
          {
	    InternalMessage internal(msg.time(), this->Processor::id(), value);
	    SingleMsgAdm::Instance().send(internal, coupled.executive());
	   }
	doneCount_m = 1;

    }
  else{
//For each message in msg bag, send External msg to all its influences.
  MessageBag::iterator bi = externalMsgs_m.begin(); //bag cursor
  for( ; bi != externalMsgs_m.end(); bi++ )
  {
    ExternalMessage xMsg(*(( ExternalMessage* )( *bi )) );
    xMsg.procId( this->Processor::id() ) ;
    //xMsg.port().print(std::cout);

    const InfluenceList &influList( xMsg.port().influences() ) ;
    InfluenceList::const_iterator pc( influList.begin() ); //port cursor

    for( ; pc != influList.end(); pc++ )
    {
      MASSERT(SingleProcessorAdmin::Instance().processor((*pc)->model()).
                        model().mType() == Model::atomicType);

      xMsg.port( *(*pc) );
      SingleMsgAdm::Instance().send( xMsg, (*pc)->model() );

      //Add procId to synchronize set if it is not there.
      syncSet_m.insert((*pc)->model());
    }//for
  }//for

  //empty msgbag
  externalMsgs_m.eraseAll();

  //printSyncSet_v();

  //send (*, t) to all syncSet members
  SyncAtomicsSet::const_iterator si;  //syncSet iterator
  for (si = syncSet_m.begin(); si != syncSet_m.end(); si++)
  {
    InternalMessage internal( msg.time(), this->Processor::id() );
    SingleMsgAdm::Instance().send( internal, *si );
  }//for

  lastChange(msg.time());

  //wait until (done, t)'s are received from all children in synchSet.
  //Then send (done, t) to the parent coordinator and clear syncSet_m
  waitingMode_m = waitingUponRecvInternalMsg_c;
  doneCount_m = syncSet_m.size();
  }

  //clear the synchronize set
  //Don't do it here. Do it in receive(Done).
  //syncSet_m.clear();

  return *this;
}//receive( const InternalMessage &msg )

/*******************************************************************
* Function Name: receive
* Description: receive Output Message; This function is called
*	       by Processor::sendOutput()
********************************************************************/
Processor &FlatDEVSCoordinator::receive( const OutputMessage &msg )
{
  DBG("OutputMessage Received in FlatDEVSCoordinator");

  const InfluenceList &influList( msg.port().influences() );
  InfluenceList::const_iterator cursor( influList.begin() );

  OutputMessage outMsg( msg );
  outMsg.procId( this->Processor::id() );

  ExternalMessage extMsg;
  extMsg.time( msg.time() );
  extMsg.procId( Processor::id() );
  extMsg.value( msg.value() );

  for( ; cursor != influList.end(); cursor++ )
  {
    //cursor is a ptr pointing to Port*
    Processor &proc(SingleProcessorAdmin::Instance().  
    			processor((*cursor)->model()));  //find by model id
    if (proc.model().mType() == Model::atomicType) 
    {
      //The dest port is an Atomic; so, treat this case as
      //if to send an External message to it
      Atomic &at( static_cast<Atomic &>(proc.model()));
      MASSERT( at.absoluteNext() >= msg.time() );
      extMsg.port( * (*cursor) );
      SingleMsgAdm::Instance().send( extMsg, (*cursor)->model() );
      syncSet_m.insert((*cursor)->model());
    }
    else if (proc.id() == this->Processor::id()) 
    {
      //The dest port is flatTop; so send outMsg to Root
      MASSERT(model().outputPorts().find((*cursor)->id()) != 
              model().outputPorts().end());
      outMsg.port( * (*cursor) );
      SingleMsgAdm::Instance().send( outMsg, model().parentId() );
    }
    else
    {
      InvalidModelIdException e ;
      e.addText( std::string( "Unexpected Model Type" ));
      MTHROW( e ) ;
    }//if-else

    //printSyncSet_v();

  }//for

  return *this; 
}//receive( OutputMessage &)

/*******************************************************************
* Function Name: receive D msg
* Description: recalculates the inminent child and sends it to his
*              father inside a done message
********************************************************************/
Processor &FlatDEVSCoordinator::receive( const DoneMessage &msg )
{
  DBG("DoneMessage Received in FlatDEVSCoordinator");
  Coupled &coupled(static_cast< Coupled &>( model() ));
  switch(waitingMode_m)
  {
    case waitingUponRecvInitMsg_c:
    {
      if( ! --doneCount_m )
      {
        waitingMode_m = noWaiting_c;

        lastChange( msg.time() );

        recalcImminentChildren();

        if( imminentChild() == Processor::InvalidId )
        {
          nextChange( Time::Inf );
        }else {
          Processor &proc = SingleProcessorAdmin::Instance().
                                 processor(*(immiChildren().begin()) );
          nextChange( proc.model().absoluteNext() - msg.time() );
        }//if-else
        DoneMessage doneMsg(msg.time(), this->Processor::id(), nextChange());
        SingleMsgAdm::Instance().send( doneMsg, model().parentId() );
      }//if
      break;
    }//case waitingUponRecvInitMsg_c

    case waitingUponRecvCollectMsg_c:
    {
      WaitForDoneQueue::const_iterator p;
      p = collectWaitForDoneQ_m.find(msg.procId());
      if (p == collectWaitForDoneQ_m.end())
      {
        InvalidMessageException e ;
        e.addText( std::string( "procId not in collectWaitForDoneQ_m!" ));
        MTHROW( e ) ;
      }else{
        collectWaitForDoneQ_m.erase(p);
      }//if-else
      if (collectWaitForDoneQ_m.size() == 0)
      {
        //We have now receive all the Ds from imminent children.
        //So, send Done to the parent.

        waitingMode_m = noWaiting_c;

        //This Done message should NOT trigger Root to advance its time,
        //because time advances should only occur during (*, t)
        //messages.
        //DoneMessage doneMsg(msg.time(), this->Processor::id(),
        //                                      model().nextChange());
        DoneMessage doneMsg(msg.time(), this->Processor::id(), Time::Zero);
        SingleMsgAdm::Instance().send( doneMsg, model().parentId() );
      }//if

      break;
    }//waitingUponRecvCollectMsg_c case

    case waitingUponRecvInternalMsg_c:
    {
      //For Debug Only: check to see if child is in synchronize set.
      SyncAtomicsSet::const_iterator p = syncSet_m.find(msg.procId());
      if (p == syncSet_m.end())
      {
        InvalidMessageException e ;
        e.addText( std::string( "procId not in syncSet_m!" ));
        MTHROW( e ) ;
      }else{
        syncSet_m.erase(p);
      }//if-else
      int value = (int)msg.value();
      if(value > 0)
	screq[msg.procId()] = value;
      if( ! --doneCount_m )
      {
        //We have received all the Done msgs from children
        lastChange( msg.time() );

        recalcImminentChildren();

        if( *(immiChildren().begin()) == Processor::InvalidId )
        {
          nextChange( Time::Inf );
        }else {
          Processor &proc = SingleProcessorAdmin::Instance().
                                 processor(*(immiChildren().begin()) );
          nextChange( proc.model().absoluteNext() - msg.time() );
        }//if-else
        if(!screq.empty())
	  {
	    SCReq::iterator it = screq.begin();
	    std::cout<<"screq is not empty the element is "<<(*it).first<<std::endl;
	    ImminentChildren::const_iterator imm_cursor = immiChildren().begin();
	    while (imm_cursor != immiChildren().end())
	      {
		if(it->first == (*imm_cursor))
		  {
		    std::cout<<"find out a structure component from imminent children "<<(*imm_cursor)<<std::endl;
		    waitingMode_m = noWaiting_c;
		    DoneMessage doneMsg(msg.time(), this->Processor::id(), coupled.nextChange(), it->second);
		    SingleMsgAdm::Instance().send(doneMsg, coupled.parentId());
		    Processor &proc = SingleProcessorAdmin::Instance().processor(*imm_cursor);
		    if(proc.model().mType() == Coupled::coupledType)
		      {
			ModelId mid = proc.model().id();
			strucset.insert(mid);
		      }
		    break;
		  }
		else imm_cursor++;
	      }
	    screq.clear();
	  }

        //V4.2: bug fix - We cannot clear syncSet_m, because it may
	//contain elements that were inserted by the Output Message 
	//Handler during the waitingUponRecvInternalMsg_c state.
	//So, instead of clear the syncSet_m when doneCount_m becomes 0,
	//we erase the element from sysncSet_m one at a time every time we 
	//receive the done msg.
        //syncSet_m.clear();
     else{
        waitingMode_m = noWaiting_c;

        DoneMessage doneMsg(msg.time(),this->Processor::id(),nextChange());
        SingleMsgAdm::Instance().send( doneMsg, model().parentId() );
         }

      }//if
      break;
    }//waitingUponRecvInternalMsg_c case
 case waitingUponRecvSelfStrucChange_c:
   {
      Processor &proc = SingleProcessorAdmin::Instance().processor(msg.procId());
      std::cout<<"flatDEVS coordinator receives a structure done message"<<std::endl;
      coupled.setremmodels();
      if(!coupled.getremmodels().empty())
	{
	  Coupled::ModelList::const_iterator rem_cur = coupled.getremmodels().begin();
	  for(; rem_cur != coupled.getremmodels().end(); rem_cur++)
	    {
	      SyncAtomicsSet::const_iterator p = syncSet_m.find(*rem_cur);
	      if(p != syncSet_m.end()) syncSet_m.erase(p);
	    }
	}
      coupled.setnewmodels();
      if((proc.model().mType() == Model::revatomicType) && (!coupled.getnewmodels().empty()))
	{
	  std::cout<<"sending start message !!"<<std::endl;
	  StartMessage startmsg(msg.time(), this->Processor::id());
	  Coupled::ModelList::const_iterator mcursor(coupled.getnewmodels().begin());
	  waitingMode_m = waitingUponRecvStartMsg_c;
	  for(; mcursor != coupled.getnewmodels().end(); mcursor++)
	    {
	      std::cout<<"model is "<<(*mcursor)<<std::endl;
	      SingleMsgAdm::Instance().send(startmsg, *mcursor);
	    }
	  doneCount_m = coupled.getnewmodels().size();
	}
      else if ((proc.model().mType() == Model::revatomicType) && (coupled.getnewmodels().empty())){
	std::cout<<"no new models, no need to send start messages !! "<<doneCount_m<<std::endl;
        if( !--doneCount_m)
	{
	  recalcImminentChildren();
	  if( *(immiChildren().begin()) == Processor::InvalidId)
	    {
	      nextChange(Time::Inf);
	    }
	  else{
	    Processor &proc = SingleProcessorAdmin::Instance().processor( *(immiChildren().begin()) );
	    nextChange(proc.model().absoluteNext() - msg.time());
	  }
	  lastChange(msg.time());
	  waitingMode_m = noWaiting_c;
	  DoneMessage doneMsg(msg.time(), this->Processor::id(), coupled.nextChange());
	  std::cout<<"next change is "<<coupled.nextChange()<<std::endl;
	  SingleMsgAdm::Instance().send(doneMsg, coupled.parentId());
      }
     
	}
      break;
     }
  case waitingUponRecvStartMsg_c:
    {
      if(!--doneCount_m)
	{
	  recalcImminentChildren();
	  if( *(immiChildren().begin()) == Processor::InvalidId )
	    {
	      nextChange( Time::Inf );
	    }
	  else{
	    Processor &proc = SingleProcessorAdmin::Instance().processor( *(immiChildren().begin()));
	    nextChange(proc.model().absoluteNext() - msg.time());
	    std::cout<<"the imminent child is "<<proc.model().description()<<" absolutenext is "<<absoluteNext()<<std::endl;
	  }
	  waitingMode_m = noWaiting_c;
	  DoneMessage doneMsg(msg.time(), this->Processor::id(), nextChange());
	  std::cout<<"next change is "<<coupled.nextChange()<<std::endl;
          std::cout<<"absolute next is "<<absoluteNext()<<std::endl;
	  SingleMsgAdm::Instance().send(doneMsg, coupled.parentId());
	}
      break;
    }
    default:
    {
      InvalidMessageException e ;
      e.addText( std::string( "Unexpected Done message!" ));
      MTHROW( e ) ;
    }//default case
  }//switch

  return *this;
}//receive DoneMessage

/*******************************************************************
* Function Name: recalcImminentChildren
* Description: Find the most imminent atomic children
********************************************************************/
FlatDEVSCoordinator& FlatDEVSCoordinator::recalcImminentChildren() 
{
  Time min( Time::Inf );
  
  ProcessorAdmin::ProcessorDB::const_iterator cursor;
  for (cursor = SingleProcessorAdmin::Instance().procDB.begin();
       cursor != SingleProcessorAdmin::Instance().procDB.end();
       cursor++)
  {
    Model &m = (cursor->second)->model();
    if (&m == NULL) continue;

    if (m.mType() == Model::atomicType)
    {
      //Cast and obtain the atomic to be initialized
      //Now we will refer to the atomic, instead of the model
      Atomic &at( static_cast<Atomic &>(m));
      //cerr << at.asString() << ": " << at.absoluteNext().asString() << endl;
      if( min > at.absoluteNext() )
      {
        min = at.absoluteNext();
      }//if
    }//if
  }//for

  immiChildren_m.clear();

  if (min == Time::Inf)
  {
    immiChildren_m.push_back(Processor::InvalidId);
    return *this;
  }//if

  for (cursor = SingleProcessorAdmin::Instance().procDB.begin();
       cursor != SingleProcessorAdmin::Instance().procDB.end();
       cursor++)
  {
    Model &m = (cursor->second)->model();
    if (&m == NULL) continue;

    if (m.mType() == Model::atomicType)
    {
      //Cast and obtain the atomic to be initialized
      //Now we will refer to the atomic, instead of the model
      Atomic &at( static_cast<Atomic &>(m));
      //cerr << at.asString() << ": " << at.absoluteNext().asString() << endl;
      if( min == at.absoluteNext() )
      {
        immiChildren_m.push_back(cursor->first);
      }//if
    }//if
  }//for

  return *this;
  
}//ecalcImminentChildren()

/*******************************************************************
* Function Name: isAtomic
* Description: Returns true in atomic models
********************************************************************/
bool FlatDEVSCoordinator::isAtomic()
{
  return false ;
}

/*******************************************************************
* Function Name: id
********************************************************************/
Processor &FlatDEVSCoordinator::id( const ProcId &pid )
{
  this->Processor::id( pid );

  MASSERT( pmodel() );
  pmodel()->id( pid );

  return *this;
}

/*******************************************************************
* Function Name: printSyncSet_v
********************************************************************/
void FlatDEVSCoordinator::printSyncSet_v()
{
  std::cout << "\nFlatDEVSCoordinator::printSyncSet_v():\n";
  SyncAtomicsSet::const_iterator si;  //syncSet iterator
  for (si = syncSet_m.begin(); si != syncSet_m.end(); si++)
  {
    std::cout << SingleProcessorAdmin::Instance().processor(*si).
                                        model().asString() << std::endl;
  }//for

  return;
}//printSyncSet_v
