/*******************************************************************
*
*  DESCRIPTION: class MainSimulator
*
* Modification History:
*
* yyyymmmdd     Name            Description
* =========     ========        ===========
* 1998Jun27     Barylko &       Initial Creation
*               Beyoglonian
* 1999Jun04     D. Rodriguez    V2.0
* 2004Apr04     Henry Yu        V3.0: Flat Coordinator with Parallel DEVS
* 2004Oct20     Henry Yu        V4.2: Changed "top" to "flattop" for flat case
* 2008sep10     M. Moallemi     V4.5: USB connection has been used to connect
*                                     to the Mindstorms Lego.  
*                                     Functions enhanced: 
*                                       MainSimulator::MainSimulator(),
*                                       MainSimulator::run(),
*                                       MainSimulator::loadExternalEvents 
*                                       and the last 6 robot controlling
*                                       functions. The previous file is
*                                       available by extension ".b"
* 2008Nov24     K. Holman       V5.0: Seperating the need for users to register
*                                     models within this file.
*
*******************************************************************/

/** include files **/
#include <iostream>
#include <sstream>
#include <fstream>

#include "mainsimu.h"   // base header
#include "ini.h"        // class Ini
#include "modeladm.h"   // class SingleModelAdm
#include "strutil.h"    // lowerCase, str2Int
#include "coupled.h"    // class Coupled
#include "zone.h"       // class zone
#include "flatcoup.h"   // class FlatCoupledCell
#include "tdcell.h"  //
#include "idcell.h"  //
#include "port.h"	//class Port
#include "debug.h"	//DBG()
#include "revatomic.h"
#include <curses.h>	//curses I/O library for keyboard control

class RevAtomic;
/** private data **/
MainSimulator *MainSimulator::instance( NULL );

/*******************************************************************
 * this variables have to be declared here in order to make the call to
 * a static function that will execute in a separate thread
 *******************************************************************/
//test
Time tnow, deadline;
/*******************************************************************/

/*******************************************************************
* Function Name: run
********************************************************************/
MainSimulator &MainSimulator::run()
{
/***************************************************************/
//  initscr();			/* Start curses mode 		*/
//  cbreak();			/* Line buffering disabled	*/
//  keypad(stdscr, TRUE);		/* We get arrow keys to control the Robot*/
//  noecho();			/* Don't echo() while we do getch */
/****************************************************************/
  std::cout<<"start to run"<<std::endl;
  if( !loader() )
  {
    MException e( "The MainSimulator loader not found!" ) ;
    e.addText( "The loader must be seted before running the simulation." ) ;
    MTHROW( e ) ;
  }
  //initialization of the the Hardware
  if (loader()->isRealTimeRun())
     InitializeHardware();
  //initialization ends

  loader()->loadData();

  DBG( "Loading Models...");
  loadModels( loader()->modelsStream(), loader()->printParserInfo() );

  loadExternalEvents( loader()->eventsStream() );//load eventfile 
  Root::Instance().stopTime( loader()->stopTime() );
  startTime_m = elapsedTime();
  Root::Instance().simulate();

  loader()->writeResults();

  if (loader()->isRealTimeRun())
	  CloseHardware();
 
  return *this;
}

/*******************************************************************
* Function Name: loadModels
********************************************************************/
MainSimulator &MainSimulator::loadModels( istream &file, bool printParserInfo )
{
  std::cout<<"start to load model"<<std::endl;
  iniFile().parse( file, MainSimulator::Instance().isFlatDEVS()?true:false);

  loadModel( Root::Instance().top(), iniFile(), printParserInfo );

  return *this;
}

/*******************************************************************
* Function Name: getParameter
********************************************************************/
const std::string MainSimulator::getParameter( const std::string &modelName, const std::string &parameterName ) const
{
  std::cout<<"modelName is "<<modelName<<" parameterName is "<<parameterName<<std::endl;
  const Ini::IdList &idList( iniFile().definition( modelName, parameterName ) );

  MASSERTMSG( !idList.empty(), "Parameter " + parameterName + " without value");

  std::string str =  idList.front();
  std::cout<<"string name is "<<str<<std::endl;
  return str;
}

/*******************************************************************
* Function Name: MainSimulator
* Description: Constructor
********************************************************************/
MainSimulator::MainSimulator() 
: separator( '@' )
, componentLabel( "components" )
, inPortsLabel("in")
, outPortsLabel("out")
, linksLabel("link")
, sloader( NULL )
, startTime_m( Time::Zero )
{
  // parselog.open("dslog.txt");
  registerNewAtomics();
  nameset.clear();
}

/*******************************************************************
* Function Name: loadModel
* Description:	loadModel uses depth-first recursion. i.e., the Atomic
*		models (leafs) are loaded first, then coupled models are
*		loaded from bottom up.
********************************************************************/
MainSimulator &MainSimulator::loadModel( Coupled &parent, Ini &ini, bool printParserInfo )
{
  // cargamos los ports
  loadPorts( parent, ini );

  // los componentes
  if( parent.type() == Coupled::cell ){
    loadCells( static_cast<CoupledCell&>( parent ), ini, printParserInfo ) ;
  }else{
    //loadComponents() calls loadModel(); i.e., depth-first recursion
    loadComponents( parent, ini, printParserInfo );	
  }//if-else

  // load links
  //if (FlatDebug().Active() && (parent.id() == Root::Instance().top().id()))
  if (MainSimulator::Instance().isFlatDEVS() && 
      (parent.id() == Root::Instance().top().id()))
  {
    loadFlattenLinks(parent, ini);
    std::cout<<"updating outlinks !!"<<std::endl;
    updateOutLinks(parent, ini);
  }else{
    loadLinks( parent, ini );
  }

  // At this time, the links are created. Now, we load the portInTransitions
  if ( parent.type() == Coupled::cell)
    loadPortInTransitions( static_cast<CoupledCell&>(parent), ini, printParserInfo ) ;

  if ((DevsMdlDebug().Active() || FlatDebug().Active()) && 
      (parent.id() == Root::Instance().top().id())) 
  {
    DevsMdlDebug().Active()?parent.print(DevsMdlDebug().Stream(), 0):
                            parent.print(FlatDebug().Stream(), 0);
  }

  return *this;
}// end MainSimulator::loadModel

/*******************************************************************
* Function Name: loadPorts
********************************************************************/
MainSimulator &MainSimulator::loadPorts( Coupled &parent, Ini &ini )
{
  std::string modelName( lowerCase( parent.description() ) ) ;

  Ini::IdList::const_iterator cursor ; 

  // ports in
  if ( ini.exists( modelName, MainSimulator::inPortsLabel ) )
  {
    const Ini::IdList &ilist( ini.definition( modelName, MainSimulator::inPortsLabel ) ) ;
    for( cursor = ilist.begin() ; cursor != ilist.end() ; cursor ++ )
      parent.addInputPort( *cursor ) ;
  }

  // ports out
  if( ini.exists( modelName, MainSimulator::outPortsLabel ) )
  {
    const Ini::IdList &ilist( ini.definition( modelName, MainSimulator::outPortsLabel ) ) ;
    for( cursor = ilist.begin() ; cursor != ilist.end() ; cursor ++ )
      parent.addOutputPort( *cursor ) ;
  }
  return *this ;
}

/*******************************************************************
* Function Name: loadComponents
********************************************************************/
MainSimulator &MainSimulator::loadComponents( Coupled &parent, Ini &ini, bool printParserInfo )
{
  std::string modelName( lowerCase( parent.description() ) );
  bool exist = ini.exists( modelName, MainSimulator::componentLabel ) ;

  if( !exist )
  {
    ini.print_v();
    MException e( std::string("Ini: Coupled Model without components!, Label: ") + modelName + " / " + MainSimulator::componentLabel ) ;
    MTHROW( e ) ;
  } else
  {
    const Ini::IdList &ilist( ini.definition( modelName, MainSimulator::componentLabel ) ) ;

    Ini::IdList::const_iterator cursor ;

    for( cursor = ilist.begin() ; cursor != ilist.end() ; cursor ++ )
    {
      int found;
      found = (*cursor).find('#');
      bool flag;
      if(found < 0) flag = false;
      else flag = true;
      if(flag)
       {
          std::cout<<"loading executive models"<<std::endl;
	  std::string mName, revatomicType;
	  bool isRevAtomic = splitString(*cursor, mName, revatomicType, '#');
	  if(isRevAtomic)
	    {
	      RevAtomic &model(SingleModelAdm::Instance().newRevAtomic(revatomicType, mName));
	      parent.addModel(model);
	    } 
	}
      else{
      std::string mName, atomicType ;

      bool isAtomic = splitString( *cursor, mName, atomicType, MainSimulator::separator ) ;

      if( isAtomic )
      {
	std::cout<<"current model is "<<mName<<" type is "<<atomicType<<std::endl;  
        Atomic &model( SingleModelAdm::Instance().newAtomic(atomicType, mName));
	std::cout<<"create an atomic model "<<mName<<std::endl;
        parent.addModel( model );
	if(parent.description() == Root::Instance().top().description())
	  AddComponent(mName);
      } else {
        // es un acoplado
        Coupled *coupled ;

        // Si es un acoplado puede ser un CoupledCell
        if( ini.exists( mName, "type" ) )
        {
          std::string type( ini.definition( mName, "type" ).front() ) ;

          if( type == "cell" )
            coupled = &( SingleModelAdm::Instance().newCoupledCell( mName ) ) ;
          else if( type == "flat" )
            coupled = &( SingleModelAdm::Instance().newFlatCoupledCell(mName));
          else
            MASSERTMSG( false, "Invalid coupled type " + type + "!" ) ;
        } else{
          coupled = &( SingleModelAdm::Instance().newCoupled( mName ) ) ;
        }//if-else

        parent.addModel( *coupled ) ;

        loadModel( *coupled, ini, printParserInfo ) ;
      }
      }
    }
  }
  return *this ;
}

/*******************************************************************
* Function Name: loadCells
********************************************************************/
MainSimulator &MainSimulator::loadCells( CoupledCell &parent, Ini &ini, bool printParserInfo )
{
  std::string modelName( lowerCase( parent.description() ) ) ;

  // get the width and height
  if ( ini.exists( modelName,  "dim" ) )
  {
    MASSERTMSG( !ini.exists( modelName, "width" ) && !ini.exists( modelName, "height" ), "The DIM clause can't coexist with the WITDH or the HEIGHT clauses");
    Ini::IdList  dimension = ini.definition( modelName, "dim" );
    CellPosition  cp( ini.join(dimension) );
    nTupla    dims( cp );

    MASSERTMSG( dims.dimension() >= 2, "The tuple defined with the DIM clause is invalid. It's must have 2 or more elements.");
    parent.dim( dims );
  }
  else
  {
    MASSERTMSG( !ini.exists( modelName, "dim" ), "The DIM clause can't coexist with the WITDH or the HEIGHT clauses");
    int width  = str2Int( ini.definition( modelName, "width"  ).front() ) ;
    int height = str2Int( ini.definition( modelName, "height" ).front() ) ;

    parent.dim( DIM_WIDTH, width );
    parent.dim( DIM_HEIGHT, height );
  }

  Real initialValue( str2Real( ini.definition( modelName, "initialvalue" ).front() ) ) ;
  parent.initialCellValue( initialValue ) ;

  bool inertial = ini.definition( modelName, "delay" ).front() == "inertial" ;
  parent.inertialDelay( inertial ) ;

  bool wrapped = ini.definition( modelName, "border" ).front() == "wrapped" ;
  parent.borderWrapped( wrapped ) ;

  int milSec( str2Int( ini.definition( modelName, "defaultDelayTime" ).front() ) ) ;
  parent.defaultDelay( Time( 0, 0, 0, milSec ) ) ;

  CellPositionList neighborsList ;

  const Ini::IdList &neighbors( ini.definition( modelName, "neighbors" ) ) ;

  for( Ini::IdList::const_iterator cursor = neighbors.begin() ; cursor != neighbors.end() ; cursor++ )
    neighborsList.push_back( CellPosition( *cursor ) ) ; 

  loadDefaultTransitions( parent, ini, printParserInfo ) ;

  CellPositionList selectList ;
  if( ini.exists( modelName,  "select" ) )
  {
    const Ini::IdList &select( ini.definition( modelName, "select" ) ) ;
    for( Ini::IdList::const_iterator cursor = select.begin() ; cursor != select.end() ; cursor++ )
      selectList.push_back( CellPosition( *cursor ) ) ; 
  }

  parent.createCells( neighborsList, selectList ) ;

  loadInitialCellValues( parent, ini ) ;

  loadLocalZones( parent, ini, printParserInfo ) ;

  return *this ;
}

/*******************************************************************
* Function Name: loadLinks
********************************************************************/
MainSimulator &MainSimulator::loadLinks( Coupled &parent, Ini &ini )
{
  std::string modelName( lowerCase( parent.description()));

  Ini::IdList::const_iterator cursor; 

  // links ( port(@model)? )
  if( ini.exists( modelName, MainSimulator::linksLabel ) )
  {
    const Ini::IdList &ilist( ini.definition( modelName, 
    				MainSimulator::linksLabel ) );
    for( cursor = ilist.begin(); cursor != ilist.end(); cursor ++ )
    {
      std::string sourceName, sourcePort, destName, destPort;

      //splitString() splits "sourcePort@sourceName". If
      //"@sourceName" is omitted, then splitString() fills in
      //sourcePort and returns 0.
      if( !splitString( *cursor, sourcePort, sourceName, 
      				MainSimulator::separator))
      {
        sourceName = modelName;
      }

      cursor++;

      // the cursor must be valid
      if( !splitString( *cursor, destPort, destName, 
      				MainSimulator::separator))
      {
        destName = modelName;
      }

      // If the model is of type CoupledCell and the input
      // port don't exist, then I create it.
      if (parent.className() != FLAT_COUPLED_CELL_NAME)
      {
        std::string modelName = SingleProcessorAdmin::Instance().
	                              processor( destName ).model().className();
        if ( modelName == TRANSPORT_DELAY_CELL_NAME || 
	     modelName == INERTIAL_DELAY_CELL_NAME ) 
        {
          // the addInPort method adds a port if not exist
          static_cast<AtomicCell*>(&SingleProcessorAdmin::Instance().
	                   processor( destName).model())->addInPort( destPort );
        }//if
      }//if

      // NOTE: If the parent is a FlatCoupledCell, 
      //       then addInfluence will create the port

      // If the model is of type CoupledCell and the output
      // port don't exist, then I create it.
      if (parent.className() != FLAT_COUPLED_CELL_NAME)
      {
        std::string modelName = SingleProcessorAdmin::Instance().
	                           processor( sourceName ).model().className();
        if ( modelName == TRANSPORT_DELAY_CELL_NAME || 
	     modelName == INERTIAL_DELAY_CELL_NAME ) 
        {
          // the addOutPort method adds a port if not exist
          static_cast<AtomicCell*>(&SingleProcessorAdmin::Instance().
	              processor(sourceName).model())->addOutPort( sourcePort );
        }//if
      }//if

      // NOTE: If the parent is a FlatCoupledCell, 
      //       then addInfluence will create the port.
      parent.addInfluence( sourceName, sourcePort, destName, destPort ) ;
    }
  }
  return *this ;
}

/*******************************************************************
* Function Name: loadFlattenLinks
********************************************************************/
MainSimulator &MainSimulator::loadFlattenLinks( Coupled &ftop, Ini &ini )
{

  MASSERT(ftop.description() == Root::Instance().top().description());

  std::string modelName( lowerCase( ftop.description()));

  Ini::IdList::const_iterator cursor; 

  // links ( port(@model)? )
  if( ini.exists( modelName, MainSimulator::linksLabel ) )
  {
    const Ini::IdList &ilist( ini.definition( modelName, 
    				MainSimulator::linksLabel ) );
    for( cursor = ilist.begin(); cursor != ilist.end(); cursor ++ )
    {
      std::string sourceName, sourcePort, destName, destPort;

      //splitString() splits "sourcePort@sourceName". If
      //"@sourceName" is omitted, then splitString() fills in
      //sourcePort and returns 0.
      if( !splitString( *cursor, sourcePort, sourceName, 
      				MainSimulator::separator))
      {
        sourceName = modelName;
      }

      cursor++;

      // the cursor must be valid
      if( !splitString( *cursor, destPort, destName, 
      				MainSimulator::separator))
      {
        destName = modelName;
      }
    
      //If "source" is flattop, then we need to make sure that 
      //the "dest" is an Atmoic model. i.e., If dest is a coupled, then we 
      //need to "flatten" it (i.e., change a@flattop->b, b->c to a@flattop->c).
      //Note, We may need to add more than one link to flattop.
      //e.g., Suppose a->b@coupled and b->c,d,e,..., then we need to add
      //all of these links to flattop.
      if (SingleProcessorAdmin::Instance().processor( sourceName ).model().
          description() == Root::Instance().top().description())
      {
        addFlattenedLinksToFlatTop (sourcePort, destName, destPort);
      }else{
        ftop.addInfluence( sourceName, sourcePort, destName, destPort ) ;
	AddLink(sourceName, sourcePort, destName, destPort);
      }
    }//for
  }//if
  return *this ;
}//loadFlattenLinks()

/*******************************************************************
* Function Name: addFlattenedLinksToFlatTop
* Add flattened link(s) of "destPort@destName" to "sourcePort@flattop"
********************************************************************/
MainSimulator &MainSimulator::addFlattenedLinksToFlatTop( 
	const std::string sourcePort, const std::string destName, 
	const std::string destPort )
{
  Model &dest(SingleProcessorAdmin::Instance().processor( destName ).model());
  Coupled &ftop(Root::Instance().top());

  if (dest.mType() == Model::atomicType)
  { //base condition
    //cerr << "Add: " <<  destPort << "@" << destName << 
    //        " To: " << sourcePort << "@" << ftop.description() << endl;
    ftop.addInfluence(ftop.description(), sourcePort, destName, destPort);
    ftop.addModel(dest);
    AddComponent(destName);
    AddLink("flattop", sourcePort, destName, destPort);
  }
  else if (dest.mType() == Model::coupledType)
  {
    InfluenceList::const_iterator cursor;
    Port &p(dest.port(destPort)); //find destPort in dest's in and out PortList

    //find all the endpoints (influences) of p and pass them to
    //addFlattenedLinksToFlatTop(), i.e., recursion
    for (cursor = p.influences().begin();
         cursor!= p.influences().end(); cursor++)
    { 
      Model &m = SingleProcessorAdmin::Instance().
                              processor((*cursor)->model()).model();
      addFlattenedLinksToFlatTop (sourcePort, m.description(), 
      						(*cursor)->name());
    }//for
    RemoveLink("flattop", sourcePort, destName, destPort);   
   
    RemoveComponent(destName);
    ftop.delModelId(dest.id());
    
  }else{
    InvalidModelIdException e ;
    e.addText( std::string( "unrecognized model type" ));
    MTHROW( e ) ;
  }//if

  return *this;
}//addFlattenedLinksToFlatTop()

/*******************************************************************
* Function Name: updateOutLinks
* Update all of atomics' outputs ports' destination ports, such that
* all the destinations are final atomic destination ports. i.e., 
* replace all the coupled dest. ports with atomic dest. ports.
* Warning: this function must be called after the model is fully loaded.
********************************************************************/
MainSimulator &MainSimulator::updateOutLinks( Coupled &ftop, Ini &ini )
{
  MASSERT(ftop.description() == Root::Instance().top().description());
  ProcessorAdmin::ProcessorDB::const_iterator cursor;
  for (cursor = SingleProcessorAdmin::Instance().procDB.begin();
       cursor != SingleProcessorAdmin::Instance().procDB.end();
       cursor++)
  {
    Model &m = (cursor->second)->model();
    if (&m == NULL) continue;
    if (m.mType() == Model::atomicType)
    {
      //Cast and obtain the atomic to be initialized
      //Now we will refer to the atomic, instead of the model
      Atomic &at( static_cast<Atomic &>(m));
      ftop.addModel(m);
      AddComponent(m.description());
      //if (FlatDebug().Active()) 
      // cerr << "\nProcess atomic: " << at.asString() << endl; 
      updateAtomicOutLinks(at);
    }//if
    else if(m.mType() == Model::coupledType)
      {
	std::cout<<"the model name is "<<m.description()<<std::endl;
	if(m.description() != Root::Instance().top().description())
	  {
	ftop.delModelId(m.id());
	RemoveComponent(m.description());
	Coupled &coupled( static_cast<Coupled &>(m) );
	MLinkList::iterator mit;
        PortList::iterator pcursor;

        for(pcursor = coupled.outputPorts().begin(); pcursor != coupled.outputPorts().end(); pcursor++)
	  {
	    std::cout<<"current port is "<<pcursor->second->name()<<std::endl;
	    mit = mlinklist.begin();
	  while(mit != mlinklist.end())
	    {
	      if((mit->sourceName == m.description()) && (mit->sourcePort == pcursor->second->name())) break;
              else    mit++;
	    }
           if(mit != mlinklist.end()) RemoveLink(mit->sourceName, mit->sourcePort, mit->destName, mit->destPort);
	  }
	  }
      }
  } // for
  return *this;
}//MainSimulator::updateOutLinks()

/*******************************************************************
* Function Name: updateAtomicOutLinks
********************************************************************/
MainSimulator &MainSimulator::updateAtomicOutLinks( Atomic &at )
{
  MASSERT(at.mType() == Model::atomicType);
  PortList::iterator cursor;
  for (cursor = at.outputPorts().begin();
       cursor != at.outputPorts().end(); cursor++)
  {
    MASSERT( cursor->second ); //cursor->second is "Port *" type
    Port &pt(*(cursor->second));
    if (pt.influences().size() != 0){
      pt.flattenInfluences(pt);
    }//if
  }//for
  
  return *this;
}//MainSimulator::updateAtomicOutLinks()

/*******************************************************************
* Function Name: loadExternalEvents
********************************************************************/
MainSimulator &MainSimulator::loadExternalEvents( istream  &fileIn )
{
  Root::Instance().initialize() ;

  istream_iterator<std::string> cursor( fileIn );
  try
  {
    while( cursor != istream_iterator<std::string>() )
    {
      // The event loading differs if this is a RealTimeRun 
      // (isRealTimeRun == true) or not. In the case of Real 
      // Time run, more values are loaded from the file
      // (such as: deadline time, associated output port)
      if ( loader()->isRealTimeRun() )
      {
        // This is a RealTimeRun
        // 1st, read the time of the event
        Time time( *cursor ) ;
        cursor ++ ;

        // RTDevs modifications
        // Here we also read the deadline time from the events file if
        // the isRealTimeRun variable is set

        // 2nd, read the time of the deadline
        Time timeDeadline( *cursor ) ;
        cursor ++ ;

        // 3rd, read the input port
        Port &port( Root::Instance().top().port( *cursor ) ) ;
        cursor ++ ;

        // 4th, read the associated output port
        // If this is a RealTimeRun (isRealTimeRun == true)
        Port &outport(Root::Instance().top().port( *cursor ) ) ;
        cursor ++ ;

        // 5th, read the port value and add external event	
        Root::Instance().addExternalEvent( time, timeDeadline, 
				port, outport, str2Real( *cursor ) ) ;
        cursor++ ;
      }
      else
      {
        // This is NOT a RealTimeRun
        // 1st, read the time of the event
        Time time( *cursor ) ;
        cursor ++ ;
        
        // 2nd, read the input port
        Port &port( Root::Instance().top().port( *cursor ) ) ;
        cursor ++ ;

        // 3rd, read the port value and add external event	
        Root::Instance().addExternalEvent( time, port, str2Real( *cursor ) ) ;
        cursor++ ;
      }//if-else
    } //while
  } catch( InvalidPortRequest &e )
  {
    e.addLocation( MEXCEPTION_LOCATION() ) ;
    throw e ;
  }
    this->showEvents( Root::Instance().events() ) ;
//}//if-else real input
  return *this;
}//loadExternalEvents

/*******************************************************************
* Function Name: loadInitialCellValues
********************************************************************/
MainSimulator &MainSimulator::loadInitialCellValues( CoupledCell &parent, Ini &ini )
{
  if( ini.exists( parent.description(), "initialRow" ) )
  {
    MASSERTMSG( parent.dimension().dimension() == 2, "The initialRow clause must only be used when the dimension is 2");

    const Ini::IdList &rowsList( ini.definition( parent.description(), "initialRow" ) ) ;

    register unsigned row = 0;
    Ini::IdList::const_iterator cursor = rowsList.begin();

    while ( cursor != rowsList.end() )
    {
      // the first value is the row number
      row = str2Int( (*cursor) ) ;
      MASSERTMSG( row < parent.dim(DIM_HEIGHT), "The number of row for initialRow is out of range. It's " + int2Str(row) + " and must be in [ 0, " + int2Str( parent.dim(DIM_HEIGHT) - 1 ) + "]" ) ;

      cursor++ ;

      // Los siguientes elementos son la descripcion de la fila
      register unsigned col = 0;

      while ( col < parent.dim( DIM_WIDTH ) )
      {
        MASSERTMSG( cursor != rowsList.end(), "Insuficient data for initialRow. Last row with " + int2Str(col) + " elements. Note: May be a middle row definition with std::less elements." ) ;
        std::string rowVal( (*cursor) );

        nTupla  t;
        t.add(row,col);

        parent.setCellValue( CellPosition(t), str2Real(rowVal) );
        col++ ;
        cursor++;
      }
    }
  }

  if( ini.exists( parent.description(), "initialRowValue" ) )
  {
    MASSERTMSG( parent.dimension().dimension() == 2, "The initialRowValue clause must only be used when the dimension is 2");

    const Ini::IdList &rowsList( ini.definition( parent.description(), "initialRowValue" ) ) ;

    register unsigned row = 0;
    Real val;

    for ( Ini::IdList::const_iterator cursor = rowsList.begin(); cursor != rowsList.end(); cursor++ )
    {
      // the first value is the row number
      row = str2Int( (*cursor) ) ;

      MASSERTMSG( row < parent.dim(DIM_HEIGHT), "The number of row for initialRowValue is out of range. It's " + int2Str(row) + " and must be in [ 0, " + int2Str( parent.dim(DIM_HEIGHT) - 1 ) + "]" ) ;

      cursor++ ;

      MASSERTMSG( cursor != rowsList.end(), "Invalid initial row value for initialRowValue (must be a pair rowNumber rowValues)!" );

      // the second is the description of the row
      std::string rowVal( *cursor ) ;      

      MASSERTMSG( rowVal.size() == parent.dim(DIM_WIDTH), "The size of the rows for the initial values of the CoupledCell must be equal to the width value !" );

      register unsigned col = 0;

      for (std::string::iterator rowCurs = rowVal.begin(); rowCurs != rowVal.end(); rowCurs++ )
      {
        if (*rowCurs >= '0' && *rowCurs <= '9')
          val.value( *rowCurs - '0');
        else
          val = Real::tundef;

        nTupla  t;
        t.add( row, col);

        parent.setCellValue( CellPosition(t), val );
        col++;
      }
    }
  }

  if( ini.exists( parent.description(), "initialCellsValue" ) )
    loadInitialCellValuesFromFile( parent, ini.definition(parent.description(), "initialCellsValue").front() );

  if( ini.exists( parent.description(), "initialMapValue" ) )
    loadInitialCellValuesFromMapFile( parent, ini.definition(parent.description(), "initialMapValue").front() );

  return *this ;
}

int getLineOnFile(FILE *fileIn, char *linea)
{
  #define  FILE_BUFFER_SIZE 25000
  static  char  buffer[FILE_BUFFER_SIZE];
  static  long  leido = FILE_BUFFER_SIZE;
  static  long  porLeer = FILE_BUFFER_SIZE;
  register  bool  encontre = false;
  register  long  pos = 0;

  while (!encontre)
  {
    if (leido == FILE_BUFFER_SIZE || porLeer == 0)
    {
      if (feof(fileIn) || (porLeer = fread( buffer, 1, FILE_BUFFER_SIZE, fileIn )) == 0)
        return 0;

      leido = 0;
    }

    if (buffer[leido] != '\n')
      linea[pos++] = buffer[leido++];
    else
    {
      encontre = true;
      leido++;
    }

    porLeer--;
  }
  linea[pos] = 0;
  return 1;
}

/*******************************************************************
* Function Name: loadInitialCellValuesFromFile
********************************************************************/
MainSimulator &MainSimulator::loadInitialCellValuesFromFile( CoupledCell &parent, const std::string &fileName )
{
  FILE  *fileIn;
  char  line[250], *posi;

  fileIn = fopen( trimSpaces(fileName), "r" );
  MASSERTMSG( fileIn != NULL, "Can't open the file '" + fileName + "' defined by the initialCellsValue clause");

  while (getLineOnFile(fileIn, line))
    if (line != NULL && (posi = strchr(line, '=')) != NULL)
      parent.setCellValue( CellPosition(line), str2Real(posi+1) );

  fclose(fileIn);  
  return *this;
}

/*******************************************************************
* Function Name: loadInitialCellValuesFromMapFile
********************************************************************/
MainSimulator &MainSimulator::loadInitialCellValuesFromMapFile( CoupledCell &parent, const std::string &fileName )
{
  FILE  *fileIn;
  char  line[250];

  fileIn = fopen( trimSpaces(fileName), "r" );
  MASSERTMSG( fileIn != NULL, "Can't open the file '" + fileName + "' defined by the initialMapValue clause");

  CellPosition  counter( parent.dimension().dimension(), 0 );
  register bool  overflow = false;

  while (!overflow)
  {
    MASSERTMSG( getLineOnFile(fileIn, line), "Insuficient data in file specified with InitialMapValue");

    parent.setCellValue( counter, str2Real(line) );

    overflow = counter.next( parent.dimension() );
  }
  fclose(fileIn);  
  return *this;
}

/*******************************************************************
* Function Name: loadDefaultTransitions
********************************************************************/
MainSimulator &MainSimulator::loadDefaultTransitions( CoupledCell &parent, Ini &ini, bool printParserInfo )
{
  if( ini.exists( parent.description(), "localTransition" ) )
  {
    parent.localTransition( ini.definition( parent.description(), "localTransition" ).front() ) ;

    this->registerTransition( parent.localTransition(), ini, printParserInfo );
  }
  return *this;
}

/*******************************************************************
* Function Name: loadPortInTransitions
********************************************************************/
MainSimulator &MainSimulator::loadPortInTransitions( CoupledCell &parent, Ini &ini, bool printParserInfo )
{
  // Only for CoupledCell and FlatCoupledCell
  if( ini.exists( parent.description(), "portintransition" ) )
  {
    const Ini::IdList &portInTranList( ini.definition( parent.description(), "portintransition" ) ) ;

    // define Cursor
    for( Ini::IdList::const_iterator cursor = portInTranList.begin() ; cursor != portInTranList.end() ; cursor++ )
    {
      std::string modelAndPortName( *cursor ) ; 
      cursor++;
      MASSERTMSG( cursor != portInTranList.end(), "Invalid portInTransition specification. Second parameter missing!" ) ;

      std::string actionName( *cursor ), sourcePort, sourceName;

      MASSERTMSG( splitString( modelAndPortName, sourcePort, sourceName, MainSimulator::separator ), "Invalid portName@cellModelName in portInTransition" );

      ////////////////////////////////////////////////////
      // get the ELSE statement
      std::string elseFunction( "" );

      if( ini.exists( actionName,  "else" ) ){
        elseFunction = ini.definition( actionName, "else" ).front() ;

        MASSERTMSG( elseFunction != actionName, "The Else Function '" + elseFunction + "' can't be the same as the portInFunction.\nProbably it will generate an infinite loop." );

        std::string  ef = elseFunction, elseEF;

        while (ef != "")
        {
          elseEF = "";
          if( ef != "" && ini.exists( ef,  "else" ) )
            elseEF = ini.definition( ef, "else" ).front() ;
          
          MASSERTMSG( ef != elseEF, "The Else Function '" + ef + "' can't be the same as the portInFunction.\nProbably it will generate an infinite loop." );
  
          registerTransitionPortIn( ef, ini, printParserInfo, elseEF );
          ef = elseEF;
        }
      }
      //////////////////////////////////////////////////////

      // parse the function definition
      registerTransitionPortIn( actionName, ini, printParserInfo, elseFunction );

      // NOTE: If the port not exist, the setPortInFunction will produce an exception.

      // Now, make an association between the function and the input port
      if (parent.className() == COUPLED_CELL_NAME)
        static_cast<AtomicCell*>(&SingleProcessorAdmin::Instance().processor(sourceName).model())->setPortInFunction( sourcePort, actionName );
      else if (parent.className() == FLAT_COUPLED_CELL_NAME)
        static_cast<FlatCoupledCell*>(&parent)->setPortInFunction( sourceName, sourcePort, actionName );
      else
        MASSERTMSG( false, "Can't set the PortIn Function in a model of type " + parent.className() );
    }
  }
  return *this ;
}

/*******************************************************************
* Function Name: loadLocalZones
********************************************************************/
MainSimulator &MainSimulator::loadLocalZones( CoupledCell &parent, Ini &ini, bool printParserInfo )
{
  if( ini.exists( parent.description(), "zone" ) )
  {
    const Ini::IdList &zoneList( ini.definition( parent.description(), "zone" ) ) ;

    // zone Cursor
    for( Ini::IdList::const_iterator zoneCurs = zoneList.begin() ; zoneCurs != zoneList.end() ; zoneCurs++ )
    {
      std::string fname( *zoneCurs ) ; 
      registerTransition( fname, ini, printParserInfo );

      zoneCurs++ ;

      MASSERTMSG( zoneCurs != zoneList.end() && *zoneCurs == "{", "Invalid zone specification, '{' missing!" ) ;
      zoneCurs++ ;

      while( zoneCurs != zoneList.end() && *zoneCurs != "}" )
      {
        Zone zone( *zoneCurs );
        for( Zone::Iterator cursor = zone.begin(); cursor != zone.end(); cursor++ )
          parent.setLocalTransition( *cursor, fname );

        zoneCurs++;
      }

      MASSERTMSG( zoneCurs != zoneList.end() && *zoneCurs == "}", "Invalid zone specification, '}' missing!" ) ;
    } // zone Cursor
  }
  return *this ;
}

/*******************************************************************
* Method: registerTransition
********************************************************************/
MainSimulator &MainSimulator::registerTransition(const LocalTransAdmin::Function &fName, Ini &ini, bool printParserInfo )
{
  if( !SingleLocalTransAdmin::Instance().isRegistered( fName ) )
  {
    if( ini.exists( fName ) && ini.exists( fName, "rule" ) )
    {
      const Ini::IdList &fnList( ini.definition( fName, "rule" ) ) ;

      std::stringstream buffer;
      for( Ini::IdList::const_iterator cursor = fnList.begin() ; cursor != fnList.end() ; cursor++ )
        buffer << " " << *cursor ;

      if( printParserInfo == true )
      {
        ParserDebug().Stream() << "********* BUFFER ********" << std::endl ;
        ParserDebug().Stream() << buffer.str() << std::endl;
      }

      SingleLocalTransAdmin::Instance().registerTransition( fName, buffer, printParserInfo, false) ;
    } else
    {
      MException e( std::string( "The local transition " ) + fName + " is missing in the ini file!" ) ;
      MTHROW( e ) ;
    }
  }
  return *this;
}

/*******************************************************************
* Method: registerTransitionPortIn
********************************************************************/
MainSimulator &MainSimulator::registerTransitionPortIn(const LocalTransAdmin::Function &fName, Ini &ini, bool printParserInfo, const std::string &elseFunction )
{
  if( !SingleLocalTransAdmin::Instance().isRegistered( fName ) )
  {
    if( ini.exists( fName ) && ini.exists( fName, "rule" ) )
    {
      const Ini::IdList &fnList( ini.definition( fName, "rule" ) ) ;

      std::stringstream buffer ;
      for( Ini::IdList::const_iterator cursor = fnList.begin() ; cursor != fnList.end() ; cursor++ )
        buffer << " " << *cursor ;

      if( printParserInfo == true )
      {
        ParserDebug().Stream() << "********* BUFFER (for PortIn Transition) ********" << std::endl ;
        ParserDebug().Stream() << buffer.str() << std::endl;
      }
      SingleLocalTransAdmin::Instance().registerTransition( fName, buffer, printParserInfo, true, elseFunction) ;
    } else
    {
      MException e( std::string( "The local transition " ) + fName + " is missing in the ini file!" ) ;
      MTHROW( e ) ;
    }
  }
  return *this;
}

/*******************************************************************
* Function Name: showEvents
********************************************************************/
void MainSimulator::showEvents( const EventList &events, ostream &out  )
{
  for (EventList::const_iterator cursor = events.begin() ; cursor != events.end() ; cursor++ )
    out << cursor->asString() << std::endl;
}

/*******************************************************************
* Method: splitString
********************************************************************/
bool MainSimulator::splitString( const std::string &full, std::string &first, std::string &second, char separator )
{
  first = second = "";
  register int index = full.find( separator );
  if( index >= 0 )
  {
    first = full.substr( 0, index );
    second = full.substr( index + 1 );
  } else
    first  = full;

  return index >= 0;
}

/*******************************************************************
* Function Name: elapsedTime
********************************************************************/

Time MainSimulator::elapsedTime() {
  return Time::currentTime() - startTime_m;
}

/*bool  MainSimulator::IsStrucComponent(std::string ModelName)
{
  Ini::StGroupList::const_iterator it = ini.stgroupList().begin();
  it = ini.stgroupList().find(ModelName);
  if(it != ini.stgroupList().end()) return true;
  else return false;
}*/

/*******************************************************************
* Function Name: AddComponent
********************************************************************/

bool MainSimulator::AddComponent(std::string mName)
{
  ComponentSet::iterator cit = nameset.find(mName);
  std::cout<<"(AddComponent) Component : "<<mName<<std::endl;
  if(cit != nameset.end()) return false;
  else {
    nameset.insert(mName);
    return true;
  }
}

bool MainSimulator::RemoveComponent(std::string mName)
{
  ComponentSet::iterator cit = nameset.find(mName);
  std::cout<<"(RemoveComponent) Component : "<<mName<<std::endl;
  if(cit == nameset.end()) return false;
  else{
    nameset.erase(mName);
    return true;
  }
}
bool MainSimulator::AddLink(std::string sourceName, std::string sourcePort, std::string destName, std::string destPort)
{
  MLinkList::iterator lit = mlinklist.begin();
  while(lit != mlinklist.end())
    {
      if((lit->sourceName == sourceName) && (lit->sourcePort == sourcePort) && (lit->destName == destName) && (lit->destPort == destPort)) break;
      else lit++;
    }
  if(lit == mlinklist.end())
    {
      MAINLINK mlink;
      mlink.sourceName = sourceName;
      mlink.sourcePort = sourcePort;
      mlink.destName = destName;
      mlink.destPort = destPort;
      mlinklist.push_back(mlink);
      std::cout<<"^^^added link is "<<mlink.sourceName<<" "<<mlink.sourcePort<<" "<<mlink.destName<<" "<<mlink.destPort<<std::endl;
      return true;
    }
  else return false;
}

bool MainSimulator::RemoveLink(std::string sourceName, std::string sourcePort, std::string destName, std::string destPort)
{
  MLinkList::iterator lit = mlinklist.begin();
  while(lit != mlinklist.end())
    {
      if((lit->sourceName == sourceName) && (lit->sourcePort == sourcePort) && (lit->destName == destName) && (lit->destPort == destPort)) break;
      else lit++;
    }
  if(lit != mlinklist.end()) 
   {
    std::cout<<"^^^removed link is "<<lit->sourceName<<" "<<lit->sourcePort<<" "<<lit->destName<<" "<<lit->destPort<<std::endl;
     mlinklist.erase(lit);
    
     return true;
   }
  else return false;
}


/*******************************************************************



  Mohammad's code starts here
  add all HW control functon here onward
  
  note: modify root.cpp if if you need advanced event generation


*******************************************************************/

#if 0
/*******************************************************************
* Function Name: spin_motor_clockwise
*******************************************************************/
const bool MainSimulator::Spin_Motor_Clockwise(int speed)
{
	//rotate forward motor on port A
	NXT::Motor::SetForward(OUT_A,speed);
	//rotate forward motor on port B
	NXT::Motor::SetForward(OUT_B,speed);
	return true; 
}

/*******************************************************************
* Function Name: spin_motor_CounterClockwise
*******************************************************************/
const bool MainSimulator::Spin_Motor_CounterClockwise(int speed)
{
	//rotate backward motor on port A
  	NXT::Motor::SetReverse(OUT_A,speed);
	//rotate backward motor on port B
  	NXT::Motor::SetReverse(OUT_B,speed);
  	return true; 
}

/*******************************************************************
* Function Name: spin_motor_Stop
*******************************************************************/
const bool MainSimulator::Spin_Motor_Stop()
{
	//Stop motor on port A
	NXT::Motor::Stop(OUT_A,true);
	//Stop motor on port B
	NXT::Motor::Stop(OUT_B,true);
	return true; 
}

/*******************************************************************
* Function Name: Turn_Vehicle_Left
*******************************************************************/
const bool MainSimulator::Turn_V_Left(int speed)
{
	//Motor on port A spins, but motor on port B floats
 	NXT::Motor::SetForward(OUT_A, speed); 
 	NXT::Motor::Stop(OUT_B,false); //false = no braking
 
  	return true; 
}

/*******************************************************************
* Function Name: Turn_Vehicle_Right
*******************************************************************/
const bool MainSimulator::Turn_V_Right(int speed)
{
	//Motor on port B spins, but motor on port A floats
 	NXT::Motor::SetForward(OUT_B, speed); 
 	NXT::Motor::Stop(OUT_A,false);

	return true;
}

/*******************************************************************
* Function Name: InitHardware
* Description: Opens hardware connections  *******************************************************************/
const bool MainSimulator::InitHardware()
{
   // Open the connection to the LEGO robot
   int open = NXT::Open();
   // Return if we cannot open the connection
   if (open == 0)
   {
      std::cout << "Could not connect:" << std::endl;
      return false;
   }
   else 
   {
      std::cout << "USB Opend" << std::endl;
   }
   // Setup: the touch sensor should be attached to input port 1
   NXT::Sensor::SetTouch(IN_1); //tell the NXT that the touch sensor is in port 1
   std::cout << "Touch sensor initialized" << std::endl;
   NXT::Sensor::SetSonar(IN_4); // tell the NXT that the sonar sensor is in port 4
   std::cout << "Sonar sensor initialized" << std::endl;
   return true;
}

/*******************************************************************
* Function Name: CloseHardware
* Description: Closes hardware connections *******************************************************************/
const bool MainSimulator::CloseHardware()
{
	//Stop motor on port A
	NXT::Motor::Stop(OUT_A,true);
	//Stop motor on port B
	NXT::Motor::Stop(OUT_B,true);
	// Close the connection to the LEGO robot
	NXT::Close(); //close the NXT

	return true;
}

/*******************************************************************
* Function Name: UpdateHardware
* Description: gets sensor values and updates sensor variabales  
*******************************************************************/
const bool MainSimulator::UpdateHardware()
{
    //Get touch sensor value;
     touchvalue = NXT::Sensor::GetValue(IN_1);
    //Retrieves the value for sonar sensors in centimeters.
     sonarvalue = NXT::Sensor::GetSonarValue(IN_4);

	return true;
}

/*******************************************************************
* Function Name: CatcheEvent
* Description: Looks in to the sensor values, if there's an event returns   * the event value
*******************************************************************/
const bool MainSimulator::CatcheEvent(int& eventvalue)		//v3
{
     //Check if touch sensor is pressed or sonar sensor detects something
        if(touchvalue || sonarvalue < 18)
	{ 
            eventvalue = touchvalue ? 1000 : sonarvalue;

            touchvalue = false; //reset touch sensor value
            sonarvalue = 0 ;//reset sonar sensor value
	    return true;
	}
    return false;
}
#endif
