/*******************************************************************
*
*  DESCRIPTION: class Model(base class for Amtomic and Coupled)
*
*  AUTHOR:    Amir Barylko & Jorge Boyoglonian 
*  Version 2: Daniel Rodriguez
*
*  EMAIL: mailto://amir@dc.uba.ar
*         mailto://jbeyoglo@dc.uba.ar
*         mailto://drodrigu@dc.uba.ar
*
*  DATE: 27/6/1998
*  DATE: 27/2/1999 (v2)
*
*******************************************************************/

#ifndef __MODEL_H
#define __MODEL_H

/** include files **/
#include <string>    // Template string
#include "port.h"     // class Port 
#include "cdTime.h"     // class Time
#include "portlist.h"
#include "procadm.h"

/** foward declarations **/

class ModelAdmin ;
class MessageAdmin ;
class Coupled ;
class Root ; 
class Simulator ;
class Processor ;
class Coordinator ;
class SendPortNode ;

/** definitions **/

#define  MODEL_NAME  "Model"

class Model
{
public:
  virtual ~Model();  // Destructor

  enum ModelType{
    atomicType = 0,
    coupledType = 1,
    revatomicType = 2

  };
  
  virtual ModelType mType() = 0;
  virtual std::string className() const = 0;

  virtual Port &addInputPort( const std::string & );
  virtual Port &addOutputPort( const std::string & );

  Model &delInputPort( const std::string & );
  Model &delOutputPort( const std::string & );

  virtual const PortList &inputPorts() const
      {return inputList;}

  virtual const PortList &outputPorts() const
      {return outputList;}

  Port & port( const std::string & );

  const Time &nextChange() const
    {return SingleProcessorAdmin::Instance().processor( id() ).nextChange();}

  const Time &lastChange() const
      {return SingleProcessorAdmin::Instance().processor( id() ).lastChange();}

  const int strucChange() const
    {return SingleProcessorAdmin::Instance().processor( id() ).strucChange();}

  const ModelId &id() const
      {return ident;}

  const std::string asString() const
      {return description() + "(" + id() + ")";}

  const Time absoluteNext() const
      {return lastChange() + nextChange();}

  virtual PortList &inputPorts()
      {return inputList;}
      
  virtual PortList &outputPorts()
      {return outputList;}

  const ModelId &parentId() const
      {return parent_id;}
  const ModelId &executive() const
    {return execId;}
      
  Model &parentId( const ModelId &mid )
      {parent_id = mid; return *this;}

  const Model &parent() const
      {return SingleProcessorAdmin::Instance().processor( parentId() ).model();}
  Model &parent()
    {return SingleProcessorAdmin::Instance().processor(parentId()).model();}
  Model &parent( const Model &p )
      {parent_id = p.id(); return *this;}

  const std::string &description() const
      {return descript;}

  bool operator <( const Model &model ) const
      {return this->id() < model.id();}

  virtual Model &print(std::ostream &os, int indent = 0); 

protected:

  friend class Processor ;
  friend class Simulator ;
  friend class RevSimulator;
  friend class Coordinator ;
  friend class SendPortNode ;
  friend class FlatDEVSCoordinator ;

  Model( const std::string &name = "Model" ) ; // Default constructor

  Model &nextChange( const Time & );
  Model &lastChange( const Time & );

  Model &strucChange(int &);

  Model &id( const ModelId &mid )
      {ident = mid; return *this;}

  Model &sendOutput( const Time &time, const Port &port, const Value &value )
      {processor().sendOutput( time, port, value ); return *this;}

  Processor &processor()
      {return SingleProcessorAdmin::Instance().processor( id() );}

private:
  friend class ModelAdmin ;
  friend class MessageAdmin ;
  friend class Coupled ;
  friend class Root ;

  ModelId ident ;
  ModelId parent_id ;
  ModelId execId;
  std::string descript ;

  PortList inputList ;
  PortList outputList ;

};  // class Model


#endif   //__MODEL_H 
