/*******************************************************************
*
*  DESCRIPTION: class Root
*
* Modification History:
*
* yyyymmmdd     Name            Description
* =========     ========        ===========
* 1998Jun27     Barylko & Beyoglonian     Initial Creation
* 1999Mar02     D. Rodriguez    V2.0
* 2004Apr04     Henry Yu        V3.0: Flat Coordinator with Parallel DEVS
* 2004Oct20     Henry Yu        V4.2: Bug fix
* 2005Jul07     Henry Yu        V4.3: Bug fix: after (x, t) is sent, should 
*				then send either (*, t) or (@, t) depending
*				on the waitingMode_m
* 2008Oct10     M. Moallemi     V4.4: Output & Done Msgs Manipulated for Robot
*
*******************************************************************/

/** include files **/
#include "evaldeb.h"  // ShowVirtualTimeWhenFinish
#include "root.h"       // base
#include "msgadm.h"     // class MessageAdm
#include "message.h"    // class Message
#include "modeladm.h"   // class SingleModelAdm
#include "coupled.h"    // class Coupled
#include "mainsimu.h"   // class Simulator
#include "debug.h"      // DBG

/** public **/
/*******************************************************************
* Function Name: Instance
* Description: 
********************************************************************/
Root &Root::Instance()
{
  static bool instance( false );

  if( !instance )
  {
    SingleProcessorAdmin::Instance().generateRoot()  ;
    instance = true ;
  }

  return static_cast<Root &>( SingleProcessorAdmin::Instance().processor( ProcessorAdmin::rootId ) ) ;
}

/*******************************************************************
* Function Name: Root
* Description: constructor Default
********************************************************************/
Root::Root( const ProcId &id )
: child( NULL )
{
  this->id( id ) ;
  
  std::string name = MainSimulator::Instance().isFlatDEVS()?"flattop":"top";
  Coupled &coupled( SingleModelAdm::Instance().newCoupled( name ));

  coupled.parentId( this->id() );

  Processor &proc( SingleProcessorAdmin::Instance().processor( coupled.id() ));
  child = &proc ;

  waitingMode_m = noWaiting_c;
}

/*******************************************************************
* Function Name: initialize
* Description: clean event List and output device
********************************************************************/
Root &Root::initialize()
{
  lastChange( Time::Zero );
  nextChange( Time::Zero );
  externalEvents.erase( externalEvents.begin(), externalEvents.end() ) ;
  deadlines.erase( deadlines.begin(), deadlines.end() ) ;
 // stopTime( Time::Inf ) ;

  return *this;
}

/*******************************************************************
* Function Name: simulate
* Description: begin the simulation
********************************************************************/
Root &Root::simulate()
{
  externalEvents.sort();
  InitMessage initMsg( MainSimulator::Instance().elapsedTime()/*Time::Zero*/, id() ) ;//V4.4
  waitingMode_m = waitingForInitDone_c;
  SingleMsgAdm::Instance().send( initMsg, top().id() ) ;
  SingleMsgAdm::Instance().run() ;
  return *this;
}

/*******************************************************************
* Function Name: stop
* Description: stop the simulation
********************************************************************/
Root &Root::stop()
{
  if (ShowVirtualTimeWhenFinish().Active())
    std::cerr << "\n" << lastChg << "\n";

  SingleMsgAdm::Instance().stop() ;
  return *this;
}

/*******************************************************************
* Function Name: receive
* Description: send the time and value of the message to the output device 

* Jeremy: output functions of a CD++ atomic will be 
* interpretted here for HW output



********************************************************************/
Processor &Root::receive( const OutputMessage &msg )
{
  if ( MainSimulator::Instance().loader()->isRealTimeRun() )
  {
    // ------------------------
    // We check if the output message is being processed on time!
    // ------------------------
    Time td = Time::Zero;
    deadlines.sort();
    DeadlineList::iterator it;
    it = deadlines.begin() ;

    bool found = false;
    std::string result = "NA";
    // Remember that the list is sorted so the 
    // first deadline is the one to be compared
    // Check if there's a pending deadline for this port
    while ( !found && it != deadlines.end() )
    {
      // If there is a pending deadline for this port...
      if ( (it->getOutport()).id() == msg.port().id() )
      {
        // Found! Take a look at the deadline and message times
        // and increment success or unsuccess counter in each case
        found = true;
        td = it->timeDeadline;

        if ( it->timeDeadline < MainSimulator::Instance().elapsedTime())
        { // Are we on time with the deadline?
          // NO, output not completed on time!
          //incUnsuccessGoalCounter();
          //deadlineMissed(td, MainSimulator::Instance().elapsedTime());
          result = " (NOT succeeded)";
        }
        else 
        {
	   // YES, output completed on time!
          //incSuccessGoalCounter();
			
	//@Jeremy	: this function sends output to the hardware:
			// case 1: output successful, deadline met
			//case 2 is above, deadline not succeeded.
	  if (MainSimulator::Instance().loader()->isRealHardRun())
	    sendOutputtoHardware(msg);
          result = " (Succeeded)";
        }//if-else

        MainSimulator::Instance().outputStream() << "Cur Time: " 
       	  << MainSimulator::Instance().elapsedTime().asString()
          << " Deadline: " << (it->timeDeadline).asString() << result
  	  << " OutPort: " << msg.port().name() 
          << " PortValue: " << Real(msg.value()) << std::endl ;

        // Finally, remove the deadline from the list
        deadlines.erase(it);
      }
      else // ( (it->getOutport()).id() == (msg->port()).id() )
      {
            //cout << "        Different!" << endl;
      } // end if ( (it->getOutport()).id() == (msg->port()).id() )
      it++;
    } // end while

	//if there is no deadline set for this event (i.e. alot of the time,
	//the below code is executed:
    if(!found)
    {
	  if (MainSimulator::Instance().loader()->isRealHardRun())
	    sendOutputtoHardware(msg);
          MainSimulator::Instance().outputStream() << "Cur Time: " 
       	    << MainSimulator::Instance().elapsedTime().asString()
            << " (no dealine specified)" 
	    << " OutPort: " << msg.port().name() 
            << " PortValue: " << Real(msg.value()) << std::endl ;
    }//if

  }//if Real-time mode
  else
  {
    MainSimulator::Instance().outputStream() 
  	<< msg.time().asString() << " " 
  	<< msg.port().name() << " " << Real(msg.value()) << std::endl ;
  }//if-else

  return *this;

}//Root::receive( const OutputMessage &msg )

/*******************************************************************
* Function Name: receive
* Description: advance the time and go on with the simulation
********************************************************************/
Processor  &Root::receive( const DoneMessage &msg )
{
  bool isBlocking = false;

  lastChg = msg.time();
  nextChange( msg.nextChange() );
  Time nt( msg.time() + nextChange() ); //i.e., msg.time + msg.nextChange
  Message *pMsg = NULL;
  Message *pXmsg = NULL;

  DBG("Root receives DONE message\n");

  if( externalEvents.empty() && nextChange()==Time::Inf )
  {
 //Mohammad Test, there is no event in event file but we need to wait for 	the rest of simulation time and catche real events
     while (MainSimulator::Instance().elapsedTime() < stopTime()) 
     {
       if(MainSimulator::Instance().loader()->isRealHardRun())
         if( getHardEvent())  //Check if there's a realtime event  
  	   return *this; 
     }//while
     stop() ;
  } 
  else 
  {
    // There are still events, or nextChange is not infinity, or both.
    Event ev( Time::Inf ) ;

    if( !externalEvents.empty() ){
      ev = externalEvents.front();
    }

    if( nextChange()==Time::Inf )//ev.time <= nt )
    {
      if ( MainSimulator::Instance().loader()->isRealTimeRun() )
      {
        //In real time mode, the elapsed time cannot be later than
        //ev.time; Otherwise, the simulator missed the event.
	//If the event is missed, we skip that event and process the
	//following one.
        while (MainSimulator::Instance().elapsedTime() > ev.time)
	{
          //Mohammad Test : bug fixed, first delete the current event
	  externalEvents.pop_front() ;
	 //End Mohammad Test
	  if( !externalEvents.empty() ){
            ev = externalEvents.front();
          }else{
	    //no more external events
	    if(nt==Time::Inf){
 //Mohammad Test, This is when we miss an event and there is no event in eventfile so we look for real events for the rest of simulation time
              while (MainSimulator::Instance().elapsedTime() < stopTime()) 
	      {
                if(MainSimulator::Instance().loader()->isRealHardRun())
		  if( getHardEvent())  //Check if there's a realtime event  
	            return *this; 
	      }//while
              stop();	
            }//if
	    return *this;	//---->
	  }//if-else
	}//while

        //ASSERT: elapsedTime <= ev.time

        // Dummy loop to wait for the ev time to arrive before sending
        // any external message
        isBlocking = false;
        while (MainSimulator::Instance().elapsedTime() < ev.time) 
	{
	  if (!isBlocking)
	  {
	    isBlocking = true;
            DBG("elapsedTime = " 
		<< MainSimulator::Instance().elapsedTime().asString() 
		<< "  ev.time = " << ev.time.asString() << std::endl);
	  }//if
          if(MainSimulator::Instance().loader()->isRealHardRun())
            if( getHardEvent())  //Check if there's a realtime event  
  	      return *this; 
	}//while

        isBlocking = false;
        DBG("elspsedTime = " 
		<< MainSimulator::Instance().elapsedTime().asString() 
		<< "  ev.time = " << ev.time.asString() << std::endl);

        //ASSERT: elapsedTime == ev.time

      }//if

      pXmsg = new ExternalMessage( 
	ev.time, id(), *ev.port, ev.value.value() );

      //V4.3 Bug Fix: miss the case that we need to send out (@, t) 
      //              rather than (*, t)
      if ((nt==ev.time)&&
          ((waitingMode_m==waitingForInitDone_c)||
           (waitingMode_m==waitingForInternalDone_c))) 
      {
        waitingMode_m = waitingForCollectDone_c;
        pMsg = new CollectMessage( ev.time, id() ) ;
	std::cout<< " @ built "<< "\n";
      }
      else
      {
        //ev.time < nt Or waitingMode_m==waitingForCollectDone_c
        waitingMode_m = waitingForInternalDone_c;
        pMsg = new InternalMessage( ev.time, id() ) ;
      }//if-else

      if ( MainSimulator::Instance().loader()->isRealTimeRun() )
      {
        //We add the DeadLine if and only if the DeadlineTime is 
	//less than infinite. If not, there is no associated deadline 
	//so we do not add the deadline
        if (ev.timeDeadline < Time::Inf) {
          //Root::Instance().addDeadline( ev.timeDeadline, *ev.port, *ev.outport);
          addDeadline( ev.timeDeadline, *ev.port, *ev.outport);
        }//if
      }//if
      externalEvents.pop_front() ;

    }
    else
    {
      //ev.time > nt, which means we don't need to send external
      //msg for ev, but need to send collect or internal msg
      
      //ASSERT: nt != Time::Inf

      if ( MainSimulator::Instance().loader()->isRealTimeRun() )
      {
        // Dummy loop to wait for the nt time to arrive
        isBlocking = false;
        while (MainSimulator::Instance().elapsedTime() < nt)
	{
	  if (!isBlocking)
	  {
	    isBlocking = true;
            DBG("elspsedTime = " << MainSimulator::Instance().elapsedTime().asString() << "  nt = " << nt.asString() << std::endl);
	  }//if
	}//while

        isBlocking = false;
        DBG("elspsedTime = " << MainSimulator::Instance().elapsedTime().asString() << "  nt = " << nt.asString() << std::endl);
      }//if

      switch (waitingMode_m)
      {
        case waitingForInitDone_c:
        case waitingForInternalDone_c:
        {
          waitingMode_m = waitingForCollectDone_c;
          pMsg = new CollectMessage( nt, id() ) ;
	  break;
        }
        case waitingForCollectDone_c:
        {
  	  waitingMode_m = waitingForInternalDone_c;
          pMsg = new InternalMessage( nt, id() ) ;
	  break;
        }
        default:
        {
          InvalidMessageException e ;
	  e.addText( std::string( "Wrong Root waitingMode_m\n" ));
	  MTHROW( e );
          break;
        }
      }//switch
    }//if-else

    if( stopTime() < pMsg->time() ){
      stop();
    }else{
      if (pXmsg){
        SingleMsgAdm::Instance().send( *pXmsg, top().id() );
        delete pXmsg;
      }//if
      SingleMsgAdm::Instance().send( *pMsg, top().id() );
    }//if-else

    delete pMsg;

  }//if-else

  return *this;
}//receive Done

/*******************************************************************
* Function Name: addExternalEvent
* Description: add external event
********************************************************************/
Root &Root::addExternalEvent( const Time &t, const Port &p, const Real &r )
{
  //  This function adds an external event when the virtual 
  //  time approach is used (NON REAL TIME APPROACH)
  //externalEvents.push_back( Event(t, &p, r) );

  externalEvents.push_back( Event(t, t, &p, &p, r) );

  return *this;
}

/*******************************************************************
* Function Name: addExternalEvent
* Description: add external event
********************************************************************/
Root &Root::addExternalEvent( const Time &t, const Time &dt, 
			      const Port &p, const Port &pout, const Real &r )
{
  // RTDevs modifications
  // Here we also send the deadline time to the event creation
  // and the output port in which we wait for the output at time
  externalEvents.push_back( Event(t, dt, &p, &pout, r) );
  return *this;
}

/*******************************************************************
* Function Name: top
********************************************************************/
Coupled &Root::top() 
{
  return static_cast<Coupled &>( child->model() );
}

/*******************************************************************
* Function Name: top
********************************************************************/
const Coupled &Root::top() const 
{
  return static_cast<const Coupled &>( child->model() ) ;
}

/*******************************************************************
* Function Name: addDeadline
* Description: add deadline
********************************************************************/
Root &Root::addDeadline( const Time &dt, const Port &p, const Port &pout )
{
    // RTDevs modifications
    // Here we also send the deadline time to the event creation
    // and the output port in which we wait for the output at time
  deadlines.push_back( Deadline(dt, &p, &pout) );

  DBG("NEW DEADLINE ADDED: " << dt.asString() << " " << pout.name() << std::endl);

  return *this;
}

/*******************************************************************
* Function Name: getHardEvent
* Description: catches real time events from hardware & sends X and * msgs 
* by Mohammad: probably does notneed t be modified.
********************************************************************/
bool Root::getHardEvent()
{
   //Message *pMsg = NULL;
   //Message *pXmsg = NULL;
   //int eventvalue;
   //Time td;

   std::cout  << "currenttime= "
              << MainSimulator::Instance().elapsedTime()
              << " Stoptime= " << stopTime() << std::endl;
   /*MainSimulator::Instance().UpdateHardware(); //will have to be modified in the mainSimu.cpp
   if( MainSimulator::Instance().CatcheEvent( eventvalue ) )
   { 
      // add an if statement hear to determine which sensor is sending info, then send message to corresponding DEVS port.
	  //dynamically defining the ports to te CD++ model to send these event commands
	  Port &port( Root::Instance().top().port("in") ) ;//Robot input port
      Port &outport(Root::Instance().top().port("out") ) ;
      std::cout<<"Event occurs @: "
	<< MainSimulator::Instance().elapsedTime()<< " "
	<< "Value: " << eventvalue <<  std::endl;

      //Make a new external message, send to a port in the DEVS model:
      pXmsg = new ExternalMessage(
	MainSimulator::Instance().elapsedTime(), id(), port, 
	eventvalue);
      waitingMode_m = waitingForInternalDone_c;
      //and make a new internal message: unique to EDSCD: with every external message, an internal message. 
      pMsg = new InternalMessage( 
        MainSimulator::Instance().elapsedTime(), id() ) ;
      if (MainSimulator::Instance().loader()->endeventTime() < Time::Inf) {//check if there's a user deadline
        td = MainSimulator::Instance().loader()->endeventTime() + 
	    MainSimulator::Instance().elapsedTime();
	addDeadline( td, port, outport);
      }//if

      if( stopTime() < pMsg->time() ){
        stop();
      }else{
        SingleMsgAdm::Instance().send( *pXmsg, top().id() );
        SingleMsgAdm::Instance().send( *pMsg, top().id() );
      }//if-else
      delete pXmsg;
      delete pMsg;
      return true;
   }
   return false;*/ return true;
}


/*******************************************************************
* Function Name: HardEvent
* Description: catches real time events from hardware & sends X and * msgs 

@Jeremy, called by the outputReceiveMessage function above,
the functions called are in mainSimu.cpp

********************************************************************/
Root &Root::sendOutputtoHardware(const OutputMessage &msg)
{
  /*
  if (msg.value() == 1){//Moving Forward
    MainSimulator::Instance().Spin_Motor_Stop();
    MainSimulator::Instance().Spin_Motor_Clockwise(30);
  }else if (msg.value() == 2){//Moving Backward
    MainSimulator::Instance().Spin_Motor_Stop();
    MainSimulator::Instance().Spin_Motor_CounterClockwise(30);
  }else if (msg.value() == 3){//Turning Left
    MainSimulator::Instance().Spin_Motor_Stop();
    MainSimulator::Instance().Turn_V_Left(20); 
  }else if (msg.value() == 4){//Turning Right
    MainSimulator::Instance().Spin_Motor_Stop();
    MainSimulator::Instance().Turn_V_Right(20); 
  }*/
  return *this;
}
